/*
 * Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include "mutt.h"
#include "mutt_curses.h"
#include "keymap.h"
#include "toolbar.h"
#include "sort.h"
#include "rfc822.h"

#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include <string.h>
#include <setjmp.h>
#include <sys/stat.h>
#include <errno.h>

#define msgno(X) *Context->v2r[X]

#define REDRAW_FULL 1
#define REDRAW_INDEX 2
#define REDRAW_CURRENT 4
#define REDRAW_MOTION 8
#define REDRAW_STATUS 16

#define HDR_OFFSET 1
#define HDR_CMD (LINES-1)
#define HDR_LEN (HDR_CMD-HDR_OFFSET-1)

#define CHECK_VCOUNT if (!vcount) { mutt_error(Context ? "There are no messages." : "No mailbox is open."); break; }

#define CURHDR Context->hdrs[*Context->v2r[current]]
#define OLDHDR Context->hdrs[*Context->v2r[oldcurrent]]

int ci_sigwinch = 0;
jmp_buf Jumpbuf;

/* The folder the user last saved to.  Used by ci_save_message() */
static char ci_last_folder[_POSIX_PATH_MAX] = { 0 };

#ifdef NCURSES_MOUSE_VERSION
TOOLBAR Main_toolbar[] = {
  { "Quit",  OP_MAIN_QUIT, 0 },
  { "Top",   OP_MAIN_FIRST_MESSAGE, 0 },
  { "Bot",   OP_MAIN_LAST_MESSAGE, 0 },
  { "NxtPg", OP_MAIN_NEXT_PAGE, 0 },
  { "PrvPg", OP_MAIN_PREVIOUS_PAGE, 0 },
  { "Del",   OP_MAIN_DELETE_MESSAGE, 0 },
  { "UnDel", OP_MAIN_UNDELETE_MESSAGE, 0 },
  { "Attch", OP_MAIN_VIEW_ATTACHMENTS, 0 },
  { "Mail",  OP_MAIN_MAIL, 0 },
  { "Reply", OP_MAIN_REPLY, 0 },
  { "Group", OP_MAIN_GROUP_REPLY, 0 },
  { "Forw",  OP_MAIN_FORWARD_MESSAGE, 0 },
  { 0, 0, 0 }
};
#endif

/*
 * force_subj is used to force printing of the subject in threading mode
 * so that the subject is displayed when threads overlap a page boundary
 *
 * Better display with 8-bit characters - pms@romance.kaist.ac.kr
 */
static void ci_generate_header (char *s, size_t l, HEADER *h, int force_subj)
{
  int len;
  int cols = option(OPTARROWCURSOR) ? COLS-2 : COLS;
  register int i, ch8bit;

  mutt_generate_header(s, l, h, force_subj);
  len = strlen(s);
  if (len < cols-1)
    sprintf(s+len, "%*.*s", cols-len-1, cols-len-1, " ");
  else {
     for(i=ch8bit=0; i<cols-1; i++) {
         if(s[i] & 0x80) ch8bit++; 
         else ch8bit = 0;
     }
     if((i-ch8bit)%2==0) s[cols-2] = 0x20;          
  }
  s[cols-1] = 0;
}

static void ci_show_last_error (void)
{
  CLEARLINE(LINES-1);
  SETCOLOR(MT_COLOR_ERROR);
  addstr(Errorbuf);
  SETCOLOR(MT_COLOR_NORMAL);
}

#ifdef METAMAIL
static int ci_invoke_metamail (HEADER *h, FILE *f)
{
  long chunk, bytes = h->length;
  FILE *fout;
  char tempfile[_POSIX_PATH_MAX], buffer[LONG_STRING];

  mutt_mktemp(tempfile);
  if ((fout = fopen(tempfile, "w")) == NULL) return -1;
  if (fseek(f, h->offset, 0) == -1) {
    fclose(fout);
    return -1;
  }
  while (bytes > 0) {
    chunk = (bytes > sizeof(buffer)) ? sizeof(buffer) : bytes;
    chunk = fread(buffer, 1, chunk, f);
    fwrite(buffer, 1, chunk, fout);
    bytes -= chunk;
  }
  fclose(fout);
  snprintf(buffer, sizeof(buffer), "%s -P -z %s", METAMAIL, tempfile);
  endwin();
  mutt_system(buffer);
  return 0;
}
#endif /* METAMAIL */

static int ci_display_message (HEADER *cur)
{
  char tempfile[_POSIX_PATH_MAX], buffer[STRING];
  STATE *st;
  int rc=0, builtin=0;

#ifdef METAMAIL
  if (cur->metamail) {
    ci_invoke_metamail(cur, Context->folder->fp);
    goto mark_it_read;
  }
#endif

  mutt_parse_mime_message(cur);

#ifdef _PGPPATH
  /* see if PGP is needed for this message.  if so, we should exit curses */
  if (cur->pgp) {
    if (cur->pgp == PGPENCRYPT && !pgp_valid_passphrase()) return 0;
    endwin();
  }
#endif

  st = mutt_new_state();
  st->displaying = 1;

  if (strcmp(Pager, "builtin") == 0) {
    mutt_mktemp(tempfile);
    if ((st->fpout = fopen(tempfile, "w")) == NULL) {
      mutt_error("Could not create temporary file!");
      mutt_free_state(&st);
      return(0);
    }
    builtin++;
  }
  else {
    if (!isendwin()) endwin();
    mutt_create_filter(Pager, &(st->fpout), 0);
  }

  mutt_make_string(buffer, sizeof(buffer), Pager_fmt, cur);
  if (builtin)
    buffer[COLS-1] = 0;
  else
    fprintf(st->fpout, "%s\n\n", buffer);

  st->fpin = Context->folder->fp;
  mutt_copy_header(Context->folder->fp, cur, st->fpout, (Context->weed ? CH_WEED : 0)|CH_DECODE);
  mutt_body_handler(cur->content, st);
  fclose(st->fpout);
  mutt_free_state(&st);

  if (builtin) {
    /* Invoke the builtin pager */
    rc = ci_simple_pager(buffer, tempfile, TRUE); /* removes file upon completion */
  }
  else {
    wait(&rc);
    keypad(stdscr, TRUE);
    if (option(OPTPROMPTAFTER)) {
      rc = ci_any_key_to_continue("Command: ");
      rc = keymap_lookup(rc, MENU_PAGER);
    }
    else
      rc = 0;
  }

#ifdef METAMAIL
mark_it_read:
#endif
  /* mark the message as read and update the new count, if needed */
  if (!cur->read) {
    cur->read = 1;
    cur->changed = 1;
    if (!cur->old) Context->new--;
  }

  return(rc);
}

static int ci_next_undeleted (int msgno)
{
  int i;

  for (i=msgno+1; i<Context->vcount; i++)
    if (! Context->hdrs[*Context->v2r[i]]->deleted)
      return(i);
  return(-1);
}

static int ci_previous_undeleted (int msgno)
{
  int i;

  for (i=msgno-1; i>=0; i--)
    if (! Context->hdrs[*Context->v2r[i]]->deleted)
      return(i);
  return(-1);
}

/* returns 1 if OK to proceed, 0 to abort */
static int confirm_func (const char *s)
{
  char msg[STRING];
  struct stat st;
  int ret=1;
  int is_mail_dir = strncmp(s, Maildir, strlen(Maildir)) == 0 ? 1 : 0;
  int is_folder=0;
  FILE *f;

  if (stat(s, &st) != -1) {
    if (is_mail_dir) {
      if ((f = fopen(s, "r")) == NULL) {
	mutt_error("Permission denied.");
	ret=0;
	goto finish;
      }
      fgets(msg, sizeof(msg), f);
      if (strncmp("From ", msg, 5) == 0) is_folder=1;
    }
    if (option(OPTCONFIRMAPPEND) || (!is_folder && option(OPTCONFIRMFILES))) {
      snprintf(msg, sizeof(msg), "Append messages to %s %s?",
	       is_folder ? "folder" : "file", s);
      if (mutt_yesorno(msg, 1) == 0) ret=0;
    }
  }
  else if (errno == ENOENT) {
    if (option(OPTCONFIRMCREATE) || (is_mail_dir && option(OPTCONFIRMFOLDERS))) {
      snprintf(msg, sizeof(msg), "Create %s %s?", is_mail_dir ? "folder" : "file", s);
      if (mutt_yesorno(msg, 1) == 0) ret=0;
    }
  }
  else {
    mutt_error("Could not stat %s!", s);
    ret=0;
  }

finish:
  CLEARLINE(LINES-1);
  return(ret);
}

/* returns 0 if the copy/save was successful, or -1 on error/abort */
static int ci_save_message (HEADER *h, int delete, int *redraw)
{
  int ch;
  char buffer[_POSIX_PATH_MAX] = "";

  *redraw = 0;
  if (h) mutt_default_savename(buffer, sizeof(buffer), h->env);
  CLEARLINE(LINES-1);
  printw("%s to mailbox (? for list): %s", delete ? "Save" : "Copy", buffer);
  refresh();
  ch = ci_getch();
  if (ch == '?') {    
    *redraw = REDRAW_FULL;
    folder_select(Maildir, buffer);
    if (buffer[0] == 0) return(-1);
  }
  else if (ch == ctrl('G')) {
    CLEARLINE(LINES-1);
    return(-1);
  }
  else if (!CI_is_return(ch)) {
    if (IsPrint(ch)) buffer[0] = 0;
    ungetch(ch);
    if (ci_get_field(delete ? "Save to mailbox: " : "Copy to mailbox: ", buffer, sizeof(buffer), CI_complete) == -1 || !buffer[0])
      return(-1);
  }
  
  /*
   * This is an undocumented feature of ELM pointed out to me by Felix von
   * Leitner <leitner@prz.fu-berlin.de>
   */
  if (strcmp(buffer, ".") == 0)
    strfcpy(buffer, ci_last_folder, sizeof(buffer));
  else
    strfcpy(ci_last_folder, buffer, sizeof(ci_last_folder));

  mutt_expand_path(buffer);

  /* check to make sure that this file is really the one the user wants */
  if (confirm_func(buffer) == 0) {
    CLEARLINE(LINES-1);
    return(-1);
  }

  mutt_save_message(h, buffer, delete);
  if (!h) *redraw = REDRAW_INDEX | REDRAW_STATUS;
  return(0);
}

static void ci_bounce_message (HEADER *h)
{
  char prompt[SHORT_STRING];
  char buf[LONG_STRING] = { 0 };
  ADDRESS *adr = 0, *padr;

  FOREVER {
    if (ci_get_field("Bounce message to: ", buf, sizeof(buf), 0) != 0 || !buf[0])
      return;
    if (mutt_parse_adrlist(&adr, buf, "@") == 0)
      break;
    mutt_free_address(&adr);
  }
  adr = mutt_expand_aliases(adr);

  buf[0] = 0;
  mutt_write_address(buf, sizeof(buf), adr);
  buf[(COLS+1)/2]=0; /* so we have a nicer display */
  snprintf(prompt, sizeof(prompt), "Bounce message to %s...?", buf);
  if (!mutt_yesorno(buf, 1)) {
    CLEARLINE(LINES-1);
    return;
  }

  buf[0] = 0;
  padr = adr;
  while (padr) {
    strcat(buf, " ");
    strcat(buf, padr->mailbox);
    if (*padr->host != '@') {
      strcat(buf, "@");
      strcat(buf, padr->host);
    }
    padr = padr->next;
  }
  mutt_free_address(&adr);

  if (Context->tagged)
    mutt_bounce_message(NULL, buf);
  else
    mutt_bounce_message(h, buf);
  mutt_error("Message%s bounced.", Context->tagged ? "s" : "");
}

static int ci_pipe_message (HEADER *h)
{
  char buffer[LONG_STRING];

  buffer[0] = 0;
  if (ci_get_field("Pipe to command: ", buffer, sizeof(buffer), 0) != 0 || !buffer[0])
    return 0;
  endwin();
  mutt_pipe_message(h, buffer);
  ci_any_key_to_continue(0);
  return 1;
}

static int ci_select_sort_method (void)
{
  int method = Sort; /* save the current method in case of abort */
  int ch;

  Sort = 0;
  while (!Sort) {
    mvaddstr(LINES-1,0, "Sort by (d)ate/(f)rom/(r)eceived/(s)ubj/(t)hreads/(u)nsorted/si(z)e?: ");
    ch = ci_getch();
    if (ch == ctrl('G') || CI_is_return(ch)) {
      Sort = method;
      CLEARLINE(LINES-1);
      return(-1);
    }
    switch (ch) {
    case 'D':
      Sort = SORTREVERSE;
    case 'd':
      Sort |= SORTDATE;
      break;
    case 'F':
      Sort = SORTREVERSE;
    case 'f':
      Sort |= SORTFROM;
      break;
    case 'R':
      Sort = SORTREVERSE;
    case 'r':
      Sort |= SORTRECEIVED;
      break;
    case 'S':
      Sort = SORTREVERSE;
    case 's':
      Sort |= SORTSUBJECT;
      break;
    case 't':
    case 'T':
      Sort = SORTTHREADS;
      break;
    case 'U':
      Sort = SORTREVERSE;
    case 'u':
      Sort |= SORTORDER;
      break;
    case 'Z':
      Sort = SORTREVERSE;
    case 'z':
      Sort |= SORTSIZE;
      break;
    }
  }
  CLEARLINE(LINES-1);
  return(Sort != method ? 0 : -1); /* no need to resort if it's the same */
}

/*
 * Return the index of the first new message, or failing that, the first
 * unread message.
 */
static int ci_first_message (void)
{
  int old = -1, i;

  for (i=0; i < Context->vcount; i++) {
    if (! Context->hdrs[*Context->v2r[i]]->read && ! Context->hdrs[*Context->v2r[i]]->deleted) {
      if (! Context->hdrs[*Context->v2r[i]]->old)
	return(i);
      else if (old == -1)
	old = i;
    }
  }
  return(old);
}

#define print_enriched_string(X) _print_enriched_string(X,1)

static void _print_enriched_string (unsigned char *s, int do_color)
{
  if (Sort != SORTTHREADS) {
    /* Don't need to do any work if we aren't using threads! */
    addstr(s);
    return;
  }
  
  while (*s) {
    if (*s < ctrl('G')) {
      if (do_color) SETCOLOR(MT_COLOR_TREE);
      while (*s && *s < ctrl('G')) {
	switch(*s) {
	case 1:
	  addch(ACS_LLCORNER);
	  break;
	case 2:
	  addch(ACS_LTEE);
	  break;
	case 3:
	  addch(ACS_HLINE);
	  break;
	case 4:
	  addch(ACS_VLINE);
	  break;
	case 5:
	  addch(' ');
	  break;
	case 6:
	  addch('>');
	  break;
	}
	s++;
      }
      if (do_color) SETCOLOR(MT_COLOR_NORMAL);
    }
    else {
      addch(*s);
      s++;
    }
  }
}

/*
 * This function handles the message index window as well as commands returned
 * from the pager (MENU_PAGER).
 */
void ci_main (void)
{
  volatile int top = 0;        /* first message on the screen */
  volatile int oldcurrent = 0; /* the last message the user visited */
  volatile int op;             /* function to execute */

  int current;                 /* the current message */
  int vcount;                  /* virtual message total */
  int redraw = REDRAW_FULL;
  int menu = MENU_MAIN;        /* which menu are we currently in? */
  int done = 0;                /* controls when to exit the "event" loop */
  int i = 0;
  int j;
  char buffer[LONG_STRING];
  char errbuf[SHORT_STRING];

#ifdef NCURSES_MOUSE_VERSION
  if (Mousemask) toolbar_init(Main_toolbar);
#endif

  if (Context->msgcount && option(OPTPOINTNEW)) {
    if ((current = ci_first_message()) == -1)
      current = Sort & SORTREVERSE ? 0 : Context->vcount - 1;
  }
  else
    current = 0;
  vcount = Context->vcount;
  Context->weed = 1; /* header weeding is on to begin with */

/* Consider newbie for Mutt - pms@romance.kaist.ac.kr */
  strcpy(Errorbuf, 
    "  ?-help, /-search, d)elete, f)orward, m)ail, r)eply, s)ave, u)ndelete, e(x)it");

  FOREVER {
    if (ci_sigwinch) {
      endwin();
      start_curses();
      ci_sigwinch = 0;
      redraw = REDRAW_FULL;
      menu = MENU_MAIN;
    }
    
    if (menu == MENU_MAIN) {
      int oldcount = (Context) ? Context->msgcount : 0;
      /* check for new mail in the mailbox */
      if (mutt_check_mailbox(Context)) {

	/* take note of the current message */
	if (oldcount)
	  oldcurrent = CURHDR->index;

	mutt_sort_headers(Context);

	/* restore the current message to the message it was pointing to */
	if (oldcount)
	  for (i=0; i < Context->vcount; i++) {
	    if (Context->hdrs[*Context->v2r[i]]->index == oldcurrent) {
	      current = i;
	      break;
	    }
	  }
	else
	  current = 0;

	/* need full redraw when we leave the pager */
	if (redraw != REDRAW_FULL) redraw = REDRAW_INDEX | REDRAW_STATUS;

	vcount = Context->vcount;
      }
      
      if (redraw == REDRAW_FULL) {
	clear();
#ifdef NCURSES_MOUSE_VERSION
	if (Mousemask) toolbar_post(Main_toolbar);
#endif
	ci_show_last_error();
	redraw = REDRAW_INDEX | REDRAW_STATUS;
      }

      if (redraw & REDRAW_STATUS) {
	mutt_status_line(buffer, sizeof(buffer));
	if (option(OPTXTERMMOUSE) || !option(OPTSTATUSONTOP)) {
	  CLEARLINE(LINES-2);
	}
	else
	  CLEARLINE(0);
	SETCOLOR(MT_COLOR_STATUS);
	printw("%-*.*s", COLS, COLS, buffer);
	SETCOLOR(MT_COLOR_NORMAL);
	redraw &= ~REDRAW_STATUS;
      }

      if (Context && Context->folder) {
	if (current >= top + HDR_LEN) {
	  while (top + HDR_LEN  - 1 < current) top += HDR_LEN;
	  redraw = REDRAW_INDEX;
	}
	else if (current < top) {
	  while (top > current) top -= HDR_LEN;
	  if (top < 0) top = 0;
	  redraw = REDRAW_INDEX;
	}
	else if (oldcurrent != current) {
	  if (!redraw) redraw = REDRAW_MOTION;
	}
      }

      if (redraw == REDRAW_INDEX) {
	for (i=top; i<top+HDR_LEN; i++) {
	  if (i < vcount) {
	    ci_generate_header(buffer, sizeof(buffer), Context->hdrs[*Context->v2r[i]], (i==top));
	    CLEARLINE(i-top+HDR_OFFSET);
	    
	    if (option(OPTARROWCURSOR))
	      move(i-top+HDR_OFFSET, 2);
	    else if (i == current)
	      SETCOLOR(MT_COLOR_INDICATOR);
	    _print_enriched_string(buffer, i!=current || option(OPTARROWCURSOR));
	    if (!option(OPTARROWCURSOR) && i == current)
	      SETCOLOR(MT_COLOR_NORMAL);
	  }
	  else
	    CLEARLINE(i-top+HDR_OFFSET);
	}
	redraw = 0;
      }
      else if (redraw == REDRAW_MOTION) {
	if (option(OPTARROWCURSOR)) {
	  ci_generate_header(buffer, sizeof(buffer), Context->hdrs[*Context->v2r[oldcurrent]], (oldcurrent==top));
	  mvaddstr(oldcurrent-top+HDR_OFFSET, 0, "  ");
	  print_enriched_string(buffer);
	}
	else {
	  /* redraw the current line in case the flags changed. */
	  CLEARLINE(oldcurrent-top+HDR_OFFSET);
	  ci_generate_header(buffer, sizeof(buffer), OLDHDR, (oldcurrent == top));
	  print_enriched_string(buffer);
		  
	  move(current-top+HDR_OFFSET, 0);
	  ci_generate_header(buffer, sizeof(buffer), CURHDR, (current == top));
	  SETCOLOR(MT_COLOR_INDICATOR);
	  _print_enriched_string(buffer, 0);
	  SETCOLOR(MT_COLOR_NORMAL);
	}
	redraw = 0;
      }
      else if (redraw == REDRAW_CURRENT) {
	if (option(OPTARROWCURSOR))
	  move(current-top+HDR_OFFSET, 2);
	else {
	  move(current-top+HDR_OFFSET, 0);
	  SETCOLOR(MT_COLOR_INDICATOR);
	}	      
	ci_generate_header(buffer, sizeof(buffer), CURHDR, (current == top));
	_print_enriched_string(buffer, option(OPTARROWCURSOR));
	if (!option(OPTARROWCURSOR)) SETCOLOR(MT_COLOR_NORMAL);
	redraw = 0;
      }
      
      if (Context) {
	oldcurrent = current;
	if (option(OPTARROWCURSOR)) {
	  move(current-top+HDR_OFFSET, 0);
	  if (Context->msgcount) {
	    SETCOLOR(MT_COLOR_INDICATOR);
	    addstr("->");
	    SETCOLOR(MT_COLOR_NORMAL);
	  }
	}
	else
	  move(current - top + HDR_OFFSET, COLS - 1);
      }
      else
	move(LINES-1, COLS-1);
      
      refresh();
      
      if (Timeout) timeout(Timeout*1000); /* milliseconds */      
      if ((i = ci_getch()) == ERR) continue;
      if (Timeout) timeout(-1); /* restore blocking operation */
	  
      ci_clear_error();
	  
#ifdef NCURSES_MOUSE_VERSION
      if (i == KEY_MOUSE) {
	MEVENT e;

	dprint(3, (debugfile, "ci_main(): detected mouse event.\n"));

	getmouse(&e);
	if (e.y == 0) {
	  /* This click occurred on the toolbar. */
	  if (e.bstate & BUTTON1_CLICKED) {
	    if ((op = toolbar_select(Main_toolbar, e.x)) == -1) continue;
	  }
	}
	else if (e.bstate & BUTTON1_CLICKED) {
	  if (e.y >= HDR_OFFSET && e.y <= HDR_OFFSET+HDR_LEN) {
	    current = e.y - HDR_OFFSET + top;
	    if (current > vcount - 1) current = oldcurrent;
	  }
	  continue;
	}
	else if (e.bstate & BUTTON1_DOUBLE_CLICKED) {
	  if (e.y >= HDR_OFFSET && e.y <= HDR_OFFSET+HDR_LEN) {
	    current = e.y - HDR_OFFSET + top;
	    if (current > vcount - 1) {
	      current = oldcurrent;
	      continue;
	    }
	    op = ci_display_message(Context->hdrs[*Context->v2r[current]]);
	    menu = MENU_PAGER;
	  }
	  continue;
	}
      }
      else
#endif /* NCURSES_MOUSE_VERSION */
	op = keymap_lookup(i, MENU_MAIN);
    } /* if (menu == MENU_MAIN) */
    
    switch (op) {
    case OP_MAIN_BOTTOM_OF_PAGE:

      CHECK_VCOUNT;

      current = top + HDR_LEN - 1;
      if (current > vcount - 1)
	current = vcount - 1;
      break;
	  
    case OP_MAIN_DELETE_PATTERN:

      CHECK_VCOUNT;

      ci_pattern_function(ACTIONDELETE, "Delete messages matching: ", CURHDR);
      redraw = REDRAW_INDEX;
      break;
      
    case OP_MAIN_ENTER_COMMAND:

      buffer[0] = 0;
      if (ci_get_field(":", buffer, sizeof(buffer), 0) != 0 || !buffer[0]) break;
      mutt_parse_rc_line(buffer, 0, errbuf, sizeof(errbuf));
      if (errbuf[0]) {
	/*
	 * since errbuf could potentially contain printf() sequences in it,
	 * we must call mutt_error() in this fashion so that vsprintf()
	 * doesn't expect more arguments that we passed
	 */
	mutt_error("%s", errbuf);
      }
      else {
	/* might have performed some action that changed the screen */
	redraw = REDRAW_FULL;
      }
      break;

#ifdef USE_POP
    case OP_MAIN_FETCH_MAIL:

      mutt_fetchPopMail();
      redraw = REDRAW_FULL;
      break;
#endif /* USE_POP */
      
    case OP_MAIN_FIRST_MESSAGE:

      CHECK_VCOUNT;

      current = 0;
      break;
      
    case OP_MAIN_HELP:

      ci_help(MENU_MAIN);
      redraw = REDRAW_FULL;
      break;
      
    case OP_MAIN_JUMP:

      CHECK_VCOUNT;

      buffer[0] = 0;
      ungetch(i);
      if (ci_get_field("Jump to message: ", buffer, sizeof(buffer), 0) != 0 || !buffer[0])
	break;
	
      if (! isdigit(buffer[0])) {
	mutt_error("Argument must be a message number.");
	break;
      }
	
      i = atoi(buffer);
      if (i > 0 && i <= Context->msgcount) {
	if (Context->hdrs[i-1]->virtual != -1)
	  current = Context->hdrs[i-1]->virtual;
	else
	  mutt_error("That message is not visible.");
      }
      else
	mutt_error("Invalid message number.");

      break;
	  
    case OP_MAIN_LAST_MESSAGE:

      CHECK_VCOUNT;

      current = vcount - 1;
      while (top + HDR_LEN < current) top += HDR_LEN;
      redraw = REDRAW_INDEX;
      break;

    case OP_MAIN_LIMIT:

      CHECK_VCOUNT;

      oldcurrent = CURHDR->index;
      if (ci_pattern_function(ACTIONLIMIT, "Limit to messages matching: ", CURHDR) == 0) {
	/* try to find what used to be the current message */
	current = -1;
	for (i=0; i<Context->vcount; i++) {
	  if (Context->hdrs[*Context->v2r[i]]->index == oldcurrent) {
	    current = i;
	    break;
	  }
	}
	if (current < 0) current = 0;
	vcount = Context->vcount;
	redraw = REDRAW_INDEX | REDRAW_STATUS;
      }
      break;	  

    case OP_MAIN_MIDDLE_OF_PAGE:

      CHECK_VCOUNT;

      i = top + HDR_LEN;
      if (i > vcount - 1) i = vcount - 1;
      current = top + (i - top) / 2;
      break;
      
    case OP_MAIN_NEXT_PAGE:

      CHECK_VCOUNT;

      if (top + HDR_LEN < vcount) {
	top += HDR_LEN;
	current = top;
	redraw = REDRAW_INDEX;
      }
      else
	mutt_error("You are on the last page.");
      break;
	  
    case OP_MAIN_PREVIOUS_PAGE:

      CHECK_VCOUNT;

      if (top > 0) {
	top -= HDR_LEN;
	if (top < 0) top  = 0;
	current = top;
	redraw = REDRAW_INDEX;
      }
      else
	mutt_error("You are on the first page.");
      break;
      
    case OP_MAIN_QUIT:

      if (mutt_close_mailbox(Context))
	done = 1;
      else
	redraw = REDRAW_FULL; /* new mail arrived? */
      break;
      
    case OP_MAIN_RECALL_MESSAGE:

      ci_send_message(0, 0, SENDPOSTPONED, (HEADER *)0);
      redraw = REDRAW_FULL;
      break;

    case OP_MAIN_REFRESH:

      clearok(stdscr, TRUE);
      redraw = REDRAW_FULL;
      break;
	  
    case OP_MAIN_SEARCH:

      CHECK_VCOUNT;
      
      if ((current = ci_search_command(current)) == -1)
	current = oldcurrent;
      break;

    case OP_MAIN_SORT_MESSAGES:

      if (ci_select_sort_method() == 0) {
	if (vcount) {
	  oldcurrent = CURHDR->index;
	  mutt_sort_headers(Context);
	  for (i=0; i<vcount; i++) {
	    if (Context->hdrs[*Context->v2r[i]]->index == i)
	      current = i;
	  }
	  redraw = REDRAW_INDEX | REDRAW_STATUS;
	}
      }
      break;
	  
    case OP_MAIN_SUBSHELL:

      CLEARLINE(LINES-1);
      endwin();
      puts("Type \"exit\" to return to Mutt.");
      fflush(stdout);
      mutt_system(Shell);
      break;      
      
    case OP_MAIN_SYNC_FOLDER:

      CHECK_VCOUNT;

      if (! mutt_was_modified(Context)) {
	mutt_error("Mailbox was not modified.");
	break;
      }
      
      strfcpy(buffer, Context->folder->name, sizeof(buffer));
      if (mutt_sync_mailbox(Context)) {
	if ((Context = mutt_open_mailbox(buffer, Context, 0)) != NULL) {
	  vcount = Context->vcount;
	  if (Context->msgcount && option(OPTPOINTNEW)) {
	    if ((current = ci_first_message()) == -1)
	      current = 0;
	  }
	  else
	    current = 0;
	}
	else {
	  vcount = 0;
	  current = 0;
	}
	redraw = REDRAW_FULL;
      }
      else
	redraw = REDRAW_FULL; /* new mail arrived? */
      break;
            
    case OP_MAIN_TAG_MESSAGE:

      CHECK_VCOUNT;

      if (CURHDR->tagged) {
	CURHDR->tagged = 0;
	Context->tagged--;
      }
      else {
	CURHDR->tagged = 1;
	Context->tagged++;
      }
      redraw = REDRAW_STATUS;
      if (current < vcount - 1)
	current++;
      else
	redraw |= REDRAW_CURRENT;
      break;
      
    case OP_MAIN_TAG_PATTERN:

      CHECK_VCOUNT;

      ci_pattern_function(ACTIONSELECT, "Tag messages matching: ", CURHDR);
      redraw = REDRAW_INDEX | REDRAW_STATUS;
      break;
	  
    case OP_MAIN_TOP_OF_PAGE:

      CHECK_VCOUNT;

      current = top;
      break;
      
    case OP_MAIN_UNDELETE_PATTERN:

      CHECK_VCOUNT;

      if (ci_pattern_function(ACTIONUNDELETE, "Undelete messages matching: ", CURHDR) == 0)
	redraw = REDRAW_INDEX;
      break;
      
    case OP_MAIN_UNTAG_PATTERN:

      CHECK_VCOUNT;

      if (ci_pattern_function(ACTIONUNSELECT, "Untag messages matching: ", CURHDR) == 0)
	redraw = REDRAW_INDEX | REDRAW_STATUS;
      break;
      
    case OP_MAIN_UNDELETE_MESSAGE:

      CHECK_VCOUNT;

      CURHDR->deleted = 0;
      if (current < vcount - 1)
	current++;
      else
	redraw = REDRAW_CURRENT;
      break;

    case OP_MAIN_VIEW_ATTACHMENTS:

      CHECK_VCOUNT;

      if (CURHDR->content->type == TYPEMULTIPART) {
	mutt_parse_mime_message(CURHDR);	
	ci_attach(CURHDR->content);
	redraw = REDRAW_FULL;
      }
      else
	mutt_error("This is not a multipart message.");
      break;
            
    case OP_MAIN_NEXT_THREAD:

      CHECK_VCOUNT;

      if ((current = mutt_next_thread(CURHDR)) < 0) {
	current = oldcurrent;
	mutt_error("No more threads.");
      }
      break;

    case OP_MAIN_PREVIOUS_THREAD:

      CHECK_VCOUNT;

      if ((current = mutt_previous_thread(CURHDR)) < 0) {
	  current = oldcurrent;
	  mutt_error("You are on the first thread.");
      }
      break;

    case OP_MAIN_EXECUTE_MACRO:

      mutt_execute_macro(MENU_MAIN, i);
      break;

    case OP_MAIN_DISPLAY_ADDRESS:
      
      CHECK_VCOUNT;
      
      buffer[0] = 0;
      mutt_write_address(buffer, sizeof(buffer), CURHDR->env->from);
      buffer[COLS-1] = 0;
      mutt_error(buffer);
      break;

      /* ----------------------------------------------------------------------
       * The following operations can be performed inside of the pager.
       */
      
    case OP_MAIN_ALIAS:

      CHECK_VCOUNT;

      ci_mkalias(CURHDR->env);
      break;
	  
    case OP_MAIN_BOUNCE_MESSAGE:

      CHECK_VCOUNT;

      ci_bounce_message(CURHDR);
      break;
	  
    case OP_MAIN_CHANGE_FOLDER:

      mvaddstr(LINES-1, 0, "Open mailbox ('?' for list): ");
      refresh();
      i = ci_getch();
      CLEARLINE(LINES-1);
      if (CI_is_return(i) || i == ctrl('G')) break;
      if (i != '?') {
        ungetch(i);
	buffer[0] = 0;
	if (ci_get_field("Open mailbox: ", buffer, sizeof(buffer), CI_complete) != 0 || !buffer[0])
	  break;
      }
      else {
	refresh();
	folder_select(Maildir, buffer);
	redraw = REDRAW_FULL;
	if (!buffer[0]) break;
      }
      mutt_expand_path(buffer);
      if (!mutt_is_valid_mailbox(buffer)) break;

      /* check to see if the current mailbox has been modified */
      if (mutt_was_modified(Context)) {
	if (!mutt_yesorno("Mailbox was modified, save changes?", 1)) {
	  if (!mutt_yesorno("Abandon changes to mailbox?", 0)) break;
	  mutt_fastclose_mailbox(Context);
	}
	else if (!mutt_close_mailbox(Context))
	  break;
      }
      
      if ((Context = mutt_open_mailbox(buffer, Context, 0)) != NULL) {
	if (Context->msgcount && option(OPTPOINTNEW)) {
	  if ((current = ci_first_message()) == -1)
	    current = Sort & SORTREVERSE ? 0 : Context->msgcount - 1;
	}
	else
	  current = 0;
	vcount = Context->vcount;
      }
      else {
	vcount = 0;
	current = 0;
      }
      ci_clear_error();
      redraw = REDRAW_FULL;
      break;
	  
    case OP_MAIN_DELETE_THREAD:

      CHECK_VCOUNT;

      mutt_do_thread(CURHDR, ACTIONDELETE);
      if ((current = ci_next_undeleted(current)) == -1)
	current = oldcurrent;
      else if (menu == MENU_PAGER) {
	op = OP_MAIN_DISPLAY_MESSAGE;
	continue;
      }
      redraw = REDRAW_INDEX;
      break;

    case OP_MAIN_DISPLAY_MESSAGE:
    case OP_MAIN_DISPLAY_HEADERS: /* don't weed the headers */

      CHECK_VCOUNT;

      /*
       * toggle the weeding of headers so that a user can press the key
       * again while reading the message.
       */
      if (op == OP_MAIN_DISPLAY_HEADERS) Context->weed = !Context->weed;
	
      op = ci_display_message(CURHDR);
      menu = MENU_PAGER;
      oldcurrent = current;
      continue;
      
    case OP_MAIN_EXIT:

      if (menu == MENU_MAIN) {
	mutt_fastclose_mailbox(Context);
	done = 1;
      }
      break;
	  
    case OP_MAIN_NEXT_UNDELETED:

      CHECK_VCOUNT;

      if (current >= vcount - 1) {
	if (menu == MENU_MAIN)
	  mutt_error("You are on the last message.");
	break;
      }
      if ((current = ci_next_undeleted(current)) == -1) {
	current = oldcurrent;
	mutt_error("No undeleted messages.");
      }
      else if (menu == MENU_PAGER) {
	op = OP_MAIN_DISPLAY_MESSAGE;
	continue;
      }
      break;
      
    case OP_MAIN_NEXT_MESSAGE:

      CHECK_VCOUNT;

      if (current >= vcount - 1) {
	if (menu == MENU_MAIN)
	  mutt_error("You are on the last message.");
	break;
      }
      current++;
      if (menu == MENU_PAGER) {
	op = OP_MAIN_DISPLAY_MESSAGE;
	continue;
      }
      break;
	  
    case OP_MAIN_PREVIOUS_UNDELETED:

      CHECK_VCOUNT;

      if (current < 1) {
	mutt_error("You are on the first message.");
	break;
      }
      if ((current = ci_previous_undeleted(current)) == -1) {
	current = oldcurrent;
	mutt_error("No undeleted messages.");
      }
      else if (menu == MENU_PAGER) {
	op = OP_MAIN_DISPLAY_MESSAGE;
	continue;
      }
      break;
	  
    case OP_MAIN_PREVIOUS_MESSAGE:

      CHECK_VCOUNT;

      if (current < 1) {
	if (menu == MENU_MAIN)
	  mutt_error("You are on the first message.");
	break;
      }
      current--;
      if (menu == MENU_PAGER) {
	op = OP_MAIN_DISPLAY_MESSAGE;
	continue;
      }
      break;
	  
    case OP_MAIN_COPY_MESSAGE:
    case OP_MAIN_SAVE_MESSAGE:

      CHECK_VCOUNT;

      if (ci_save_message(Context->tagged ? NULL : CURHDR,
			  (op == OP_MAIN_SAVE_MESSAGE), &redraw) != 0 ||
	  op != OP_MAIN_SAVE_MESSAGE)
	break;
      /* redraw will be set if there were tagged messages */
      if (!redraw || menu == MENU_PAGER) {
	if ((current = ci_next_undeleted(current)) == -1) {
	  current = oldcurrent;
	  if (!redraw) redraw = REDRAW_CURRENT;
	}
	else if (menu == MENU_PAGER) {
	  op = OP_MAIN_DISPLAY_MESSAGE;
	  continue;
	}
      }
      break;

    case OP_MAIN_DELETE_MESSAGE:

      CHECK_VCOUNT;

      CURHDR->deleted = 1;
      if ((current = ci_next_undeleted(current)) == -1) {
	current = oldcurrent;
	if (menu == MENU_MAIN) redraw = REDRAW_CURRENT;
      }
      else if (menu == MENU_PAGER) {
	op = OP_MAIN_DISPLAY_MESSAGE;
	continue;
      }
      break;
      
    case OP_MAIN_FORWARD_MESSAGE:

      CHECK_VCOUNT;

      ci_send_message(0, 0, SENDFORWARD, Context->tagged ? (HEADER *)0 : CURHDR);
      redraw = REDRAW_FULL;
      break;
      
    case OP_MAIN_REPLY:

      CHECK_VCOUNT;

      ci_send_message(0, 0, SENDREPLY, Context->tagged ? (HEADER *)0 : CURHDR);
      redraw = REDRAW_FULL;
      break;
      
    case OP_MAIN_GROUP_REPLY:

      CHECK_VCOUNT;

      ci_send_message(0, 0, SENDREPLY|SENDGROUPREPLY, Context->tagged ? (HEADER *)0 : CURHDR);
      redraw = REDRAW_FULL;
      break;
      
    case OP_MAIN_LIST_REPLY:

      CHECK_VCOUNT;

      ci_send_message(0, 0, SENDREPLY|SENDLISTREPLY, Context->tagged ? (HEADER *)0 : CURHDR);
      redraw = REDRAW_FULL;
      break;
      
    case OP_MAIN_MAIL:

      ci_send_message(0, 0, 0, (HEADER *)0);
      redraw = REDRAW_FULL;
      break;
      
    case OP_MAIN_NEXT_NEW:
    case OP_MAIN_NEXT_UNREAD:

      CHECK_VCOUNT;

      current = -1;
      for (i=current+1, j=0; j!=vcount; i++, j++) {
	if (i >  vcount - 1) {
	  mutt_error("Search wrapped around top.");
	  i = 0;
	}
	if (! Context->hdrs[*Context->v2r[i]]->deleted &&
	    ! Context->hdrs[*Context->v2r[i]]->read)
	{
	  if (op == OP_MAIN_NEXT_UNREAD || ! Context->hdrs[*Context->v2r[i]]->old) {
	    current = i;
	    break;
	  }
	}
      }
      if (current == -1) {
	current = oldcurrent;
	mutt_error("No new messages.");
      }
      break;

    case OP_MAIN_PIPE_MESSAGE:

      CHECK_VCOUNT;

      if (ci_pipe_message(CURHDR)) redraw = REDRAW_FULL;
      break;
      
    case OP_MAIN_PRINT_MESSAGE:

      CHECK_VCOUNT;

      if (!option(OPTASKPRINT) ||
	  (option(OPTASKPRINT) && mutt_yesorno("Print message?", 0)))
	mutt_print_message(CURHDR);
      break;
      
    case OP_MAIN_FLAG_MESSAGE:
#define TOGGLE_FLAG(x) (x)->flagged = !(x)->flagged; (x)->changed = 1

      CHECK_VCOUNT;

      TOGGLE_FLAG(Context->hdrs[*Context->v2r[current]]);
      if (menu == MENU_MAIN) redraw = REDRAW_CURRENT;
      break;
      
    case OP_MAIN_MARK_AS_NEW:

      CHECK_VCOUNT;

      if (CURHDR->read || CURHDR->old) {
	CURHDR->read = 0;
	CURHDR->old = 0;
	CURHDR->changed = 1;
	Context->new++;
	if (menu == MENU_MAIN) redraw = REDRAW_CURRENT|REDRAW_STATUS;
      }
      break;
      
#ifdef _PGPPATH
    case OP_MAIN_FORGET_PASSPHRASE:

      pgp_void_passphrase();
      mutt_error("PGP passphrase forgotten.");
      break;
#endif /* _PGPPATH */

    case OP_MAIN_BROWSE_URL:

      CHECK_VCOUNT;

      ci_browse_url(CURHDR);
      redraw = REDRAW_FULL;
      break;

    default:

      if (menu == MENU_MAIN) {
	mutt_error("Key is not bound.  Press '?' for help.");
	beep();
      }
    }
    
    if (menu == MENU_PAGER) {
      menu = MENU_MAIN;
      Context->weed = 1; /* turn header weeding back on. */
      redraw = REDRAW_FULL;
    }
    
    if (done) break;
  }
}
