/*
 * Copyright (C) 1996 Michael R. Elkins <me@cs.hmc.edu>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include "muttlib.h"
#include "sort.h"
#include "rfc_gen.h"

#include <fcntl.h>
#include <dirent.h>
#include <sys/stat.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <utime.h>

#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif

#ifdef USE_FLOCK
#include <sys/file.h>
#endif /* pms@romance.kaist.ac.kr */

#define MAXLOCKATTEMPT 5

extern void ci_endwin (const char *);

/* Determine the type of a folder */
static int get_magic (const char *p)
{
  struct stat sb;

  if (stat (p, &sb) == -1) {
    if (errno == ENOENT) {
      dprint(1,(debugfile, "get_magic: use default magic for non-existent folder %s.\n", p));
      return(Default_magic);
    }
    dprint (1, (debugfile, "get_magic: stat() returned errno = %d.\n", errno));
    return (-1);
  }

  if (S_ISREG(sb.st_mode)) {
    FILE *fp;
    char buffer[LONG_STRING];

    if ((fp = fopen(p, "r")) == NULL) return(-1);

    if (fgets(buffer, LONG_STRING, fp) == NULL) {
      /* This is either MMDF or MBOX.  Default to MBOX. */
      fclose(fp);
      return MAGICMBOX;
    }
    fclose(fp);
    if (strcmp("\001\001\001\001\n", buffer) == 0)
      return(MAGICMMDF);
    else if (strncmp("From ", buffer, sizeof("From ")-1) == 0)
      return(MAGICMBOX);
    dprint(1, (debugfile, "get_magic: unknown mailbox type: %s", buffer));
  }
  return(-1);
}

FOLDER *folder_open (const char *f, const char *mode)
{
  FOLDER *t;
  short magic;

#ifdef USE_SETGID
  /* Check to make sure that the user has permission to open this file! */
  if (*mode == 'r') {
    if (access(f, R_OK) == -1) {
      if (errno == EACCES)
	mutt_error("You don't have permission to read that mailbox!");
      else {
	mutt_error("access() returned -1 (code %d)", errno);
	dprint(1, (debugfile, "folder_open(): access() errno=%d.\n", errno));
      }
      return NULL;
    }
  }
  else if (*mode == 'w' || *mode == 'a') {
    if (access(f, F_OK) == -1 || access(f, W_OK) == -1) {
      if (errno != ENOENT) {
	mutt_error("You don't have permission to write to that mailbox!");
	return NULL;
      }
    }
  }
#endif /* USE_SETGID */

  magic = get_magic(f);
  if (magic < 0) return(0); /* get_magic prints its own error message. */
  t = (FOLDER *)safe_malloc(sizeof(FOLDER));
  t->magic = magic;
  if ((t->fp = fopen(f, mode)) == 0) {
    mutt_error("Could not open %s!", f);
    dprint(1, (debugfile, "folder_open: can't open %s (errno %d).\n", f, errno));
    safe_free((void **)&t);
    return(0);
  }
  t->name = safe_strdup(f);
  return(t);
}

void folder_close (FOLDER **f)
{
  safe_free((void **)&(*f)->name);
  fclose((*f)->fp);
  safe_free((void **)f);
}

static int copy_message (FOLDER *in, HEADER *hdr, FOLDER *out)
{
  mutt_copy_header(in->fp, hdr, out->fp, CH_FROM|CH_UPDATE);

  /* Copy the body of the message to the target destination. */
  mutt_copy_bytes(in->fp, out->fp, hdr->length - (ftell(in->fp) - hdr->offset));

  /* add the message separator */
  fputc('\n', out->fp);

  if (ferror(out->fp) != 0) {
    dprint(1, (debugfile, "copy_message(): ferror() returned nonzero!\n"));
    return(-1);
  }
  return(0);
}

int lock_file (FOLDER *f, int excl)
{
#ifdef USE_FCNTL
  struct flock lck;
#endif

#ifdef USE_DOTLOCK
  char lockfile[_POSIX_PATH_MAX], nfslockfile[_POSIX_PATH_MAX];
  int fd, count=0;
  struct stat sb;

#if !defined(WORLD_WRITABLE_SPOOL) && !defined(USE_SETGID) && !defined(HOMESPOOL)
  if (strcmp(Spoolfile, f->name) != 0)
#endif
  {
    sprintf(lockfile, "%s.lock", f->name);
  try_again:
    while ((fd = open(lockfile, O_WRONLY|O_EXCL|O_CREAT, 0644)) < 0) {
      if (errno != EEXIST && errno != EAGAIN)
	return(-1);
      if (++count == MAXLOCKATTEMPT) {
	if (mutt_yesorno("Lock count exceeded, remove lock?", TRUE)) {
	  unlink(lockfile);
	  count=0;
	  continue;
	}
	else {
	  dprint(1,(debugfile,"lock_file(): lock count exceded on %s.\n",
		    f->name));
	  return(-1);
	}
      }
      mutt_error("Waiting for lock attempt #%d...", count);
      sleep(2);
    }
    close(fd);

    /*
     * locking does not work well over NFS on many systems, so do an extra
     * check.
     */
    sprintf(nfslockfile, "%s.%s.%d", lockfile, Hostname, (int)getpid());
    link(lockfile, nfslockfile);
    if (stat(nfslockfile, &sb) == -1) {
      dprint(1,(debugfile,"lock_file(): could not stat the NFS lockfile %s.\n",
		nfslockfile));
      mutt_error("Could not stat lockfile!");
      return(-1);
    }
    unlink(nfslockfile);
    if (sb.st_nlink != 2) goto try_again;
  }
#endif /* USE_DOTLOCK */

#ifdef USE_FCNTL
  memset(&lck, 0, sizeof(struct flock));
  lck.l_type = excl ? F_WRLCK : F_RDLCK;
  lck.l_whence = SEEK_SET;
  if (fcntl(fileno(f->fp), F_SETLKW, &lck) == -1) {
    dprint(1,(debugfile, "lock_file(): fcntl errno %d.\n", errno));
    return(-1);
  }
#endif

#ifdef USE_FLOCK
  if (flock(fileno(f->fp), excl ? LOCK_EX : LOCK_SH) == -1)
    return(-1);
#endif /* USE_FLOCK */

  return(0);
}

void unlock_file (FOLDER *f)
{
#ifdef USE_FCNTL
  struct flock unlockit = { F_UNLCK, 0, 0, 0 };
#endif
#ifdef USE_DOTLOCK
  char lockfile[_POSIX_PATH_MAX];
#endif

#ifdef USE_FLOCK
  flock(fileno(f->fp), LOCK_UN);
#endif

#ifdef USE_FCNTL
  memset(&unlockit, 0, sizeof(struct flock));
  unlockit.l_type = F_UNLCK;
  unlockit.l_whence = SEEK_SET;
  fcntl(fileno(f->fp), F_SETLKW, &unlockit);
#endif

#ifdef USE_DOTLOCK
  sprintf(lockfile, "%s.lock", f->name);
  unlink(lockfile);
#endif
}

static void emergency_exit (void)
{
  ci_endwin("Error!  Mailbox was externally modified!  Exiting...");
  exit(1);
}

/*
 * This routine checks the on-disk state of the currently opened mailbox.  If
 * the mailbox has been modified, a sanity check is performed to make sure
 * that the state in-memory is still valid, such as if new mail is appended
 * to the mailbox.
 */
int mutt_check_mailbox (CONTEXT *ctx)
{
  struct stat st;
  char buffer[LONG_STRING];

  if (!ctx) return(0);

  if (stat(ctx->folder->name, &st) == -1) {
    dprint(1,(debugfile, "check_mailbox(): stat(%s, ...) returned -1 (errno=%d)\n", ctx->folder->name, errno));
    emergency_exit();
  }

  if (st.st_mtime == ctx->folder->mtime) return(0); /* Not modified... */

  dprint(1, (debugfile, "check_mailbox(): folder was modified, performing sanity check...\n"));

  if (st.st_size < ctx->folder->size) {
    dprint(1, (debugfile, "check_mailbox(): folder was smaller than when previously read!\nprevious size = %ld\ncurrent size = %ld\n", ctx->folder->size, (long)st.st_size));
    emergency_exit();
  }
  else if (st.st_size == 0) {
    /* someone touched the file, but it is still zero length, so just exit */
    ctx->folder->mtime = st.st_mtime;
    return(0);
  }

  /* lock the file if it isn't already */
  if (!ctx->locked) {
    mutt_block_signals();
    if (lock_file(ctx->folder, FALSE) == -1) {
      mutt_error("Couldn't lock mailbox %s!", ctx->folder->name);
      sleep(1);
      emergency_exit();
    }
  }

  /*
   * Check to make sure that the only change to the mailbox is that 
   * message(s) were appended to this file.  My heuristic is that we should
   * see the message separator at *exactly* what used to be the end of the
   * folder.
   */
  fseek(ctx->folder->fp, ctx->folder->size, 0);
  if (fgets(buffer, sizeof(buffer), ctx->folder->fp) == NULL) {
    unlock_file(ctx->folder);
    dprint(1,(debugfile,"check_mailbox(): fgets() returned 0!\n"));
    emergency_exit();
  }

  switch(ctx->folder->magic) {
  case MAGICMBOX:
    if (strncmp("From ", buffer, 5) != 0) {
      unlock_file(ctx->folder);
      dprint(1, (debugfile, "check_mailbox(): folder was externally rewritten!\nHere's what I got:\n%s\n", buffer));
      emergency_exit();
    }
    break;
  case MAGICMMDF:
    if (strcmp("\001\001\001\001\n", buffer) != 0) {
      unlock_file(ctx->folder);
      dprint(1, (debugfile, "check_mailbox(): folder was externally rewritten!\n"));
      emergency_exit();
    }
    break;
  }

  fseek(ctx->folder->fp, ctx->folder->size, 0);
  mutt_parse_mailbox(ctx);

  /*
   * Only unlock the folder if it was locked inside of this routine.  It may
   * have been locked elsewhere, like in sync_mailbox().
   */
  if (!ctx->locked) {
    unlock_file(ctx->folder);
    mutt_unblock_signals();
  }

  return(1); /* signal the fact that new mail has arrived. */
}

/* Close and abort any in-memory changes to the folder. */
void mutt_fastclose_mailbox (CONTEXT *ctx)
{
  int i;

  folder_close(&ctx->folder);

  for (i = 0; i < ctx->msgcount; i++) {
    mutt_free_header(&ctx->hdrs[i]);
    safe_free((void **)&ctx->v2r[i]);
  }
  safe_free((void **)&ctx->hdrs);
  safe_free((void **)&ctx->v2r);
  memset(ctx, 0, sizeof(CONTEXT));
}

/* return 1 if the folder has been modified */
int mutt_was_modified (CONTEXT *ctx)
{
  int i;

  for (i=0; i<ctx->msgcount; i++) {
    if (ctx->hdrs[i]->changed || ctx->hdrs[i]->deleted)
      return 1;
  }
  return 0;
}

int mutt_sync_mailbox (CONTEXT *ctx)
{
  FOLDER tmp;
  char tempfile[_POSIX_PATH_MAX];
  int i, save_sort = SORTORDER, saved=0, is_spool;
  int need_sort = 0; /* flag to restort mailbox if new mail arrives */
  struct stat statbuf;
  struct utimbuf utimebuf;

  dprint(1,(debugfile,"mutt_sync_mailbox(): entering\n"));

  if (access(ctx->folder->name, W_OK) == -1) {
    dprint(1,(debugfile,"mutt_sync_mailbox(): access() return -1 (errno %d).\n",
	      errno));
    mutt_unblock_signals(); /* in case we came from mutt_close_mailbox() */
    if (mutt_yesorno("Mailbox is read-only.  Abort changes?", 0)) {
      mutt_fastclose_mailbox(ctx);
      return(1);
    }
    return(0);
  }

  is_spool = (strcmp(Spoolfile, ctx->folder->name) == 0);

  /*
   * If the messages were sorted in some way other than mailbox-order,
   * resort so that we save to disk in a correct way.
   */
  if (Sort != SORTORDER) {
    save_sort = Sort;
    Sort = SORTORDER;
    mutt_sort_headers(ctx);
  }

  /*
   * need to open the file for writing in such a way that it does not truncate
   * the file, so use read-write mode.
   */
  if ((ctx->folder->fp = freopen(ctx->folder->name, "r+", ctx->folder->fp)) == NULL) {
    mutt_unblock_signals();
    mutt_error("Could not reopen mailbox!");
    dprint(1,(debugfile,"mutt_sync_mailbox(): could not reopen folder in read/write (r+) mode!\n"));
    return(0);
  }

  mutt_block_signals();

  if (lock_file(ctx->folder, TRUE) == -1) {
    dprint(1,(debugfile,"mutt_sync_mailbox(): unable to lock mailbox %s.\n",
	      ctx->folder->name));
    mutt_error("Unable to lock mailbox!");
    goto bail;
  }

  /* indicate that we have this mailbox locked */
  ctx->locked = 1;

  /* Check to make sure that the file hasn't changed on disk */
  if (mutt_check_mailbox(ctx)) {
    dprint(1,(debugfile, "mutt_sync_mailbox(): new mail received while trying to lock mailbox, aborting...\n"));
    need_sort = 1;
    mutt_error("New mail received while trying to lock folder.");
    goto bail;
  }

  /* Create a temporary file to write the new version of the mailbox in. */
#if !defined(WORLD_WRITABLE_SPOOL) && !defined(USE_SETGID)
  if (is_spool) {
    /*
     * Some systems, like Convex, have /bin/mail setuid root with the mail
     * spool directory mode 755, which means that the user can not write
     * anything (such as a temp file) except their one existing folder.
     */
    int tfd;
    
    mutt_mktemp(tempfile);
    if ((tfd = open(tempfile, O_WRONLY|O_EXCL|O_CREAT, 0600)) < 0) {
      dprint(1, (debugfile, "mutt_sync_mailbox(): could not create temp file! (errno=%d)\n", errno));
      goto bail;
    }
    tmp.fp = fdopen(tfd, "w");
  }
  else
#endif
    {
      snprintf(tempfile, sizeof(tempfile), "%s.XXXXXX", ctx->folder->name);
      mktemp(tempfile);
      tmp.fp = fopen(tempfile, "w");
    }

  if (tmp.fp == NULL) {
    dprint(1, (debugfile, "mutt_sync_mailbox(): %s: could not open temp file (errno=%d).\n",
	       tempfile, errno));
    mutt_error("Could not create temporary file!");
    goto bail;
  }

  tmp.magic = ctx->folder->magic;
  tmp.name = tempfile;
  
  if (!WriteMsginc) mutt_error("Synchronizing %s...", ctx->folder->name);

  for (i=0; i < ctx->msgcount; i++) {
    if (! ctx->hdrs[i]->deleted) {
      saved++; /* keep track of how many messages got copied. */
      if (WriteMsginc && (saved % WriteMsginc == 0 || saved == 1))
	mutt_error("Writing message %d...", saved);
      if (copy_message(ctx->folder, ctx->hdrs[i], &tmp) == -1) {
	fclose(tmp.fp);
	mutt_unlink(tempfile);
	mutt_error("Error while writing mailbox!");
	goto bail;
      }
    }
  }

  if (fclose(tmp.fp) != 0) {
    dprint(1, (debugfile, "mutt_sync_mailbox(): fclose() returned non-zero.\n"));
    unlink(tempfile);
    goto bail;
  }

  if (saved || option(OPTKEEPEMPTY) || is_spool) {
    FILE *src = fopen(tempfile, "r");
    long len;
    char buf[LONG_STRING];
    
    /* Save the state of this folder. */
    stat(ctx->folder->name, &statbuf);
    
    /*
     * Copy the folder back into place.  Note that since the stream is
     * read-write at this point, there is no need to worry about running
     * out of space since this just writes over the existing file.  However,
     * we must use ftruncate() to free up any extra space when done
     * copying the temporary file.
     */
    fseek(ctx->folder->fp, 0, 0);
    while ((len = fread(buf, 1, sizeof(buf), src)) > 0)
      fwrite(buf, 1, len, ctx->folder->fp);
    ftruncate(fileno(ctx->folder->fp), ftell(ctx->folder->fp));
    
    /*
     * Need to call unlock_file() instead of just closing the stream because
     * there could be a .lock file which wouldn't be removed when closing
     * the stream like the fcntl() lock is.
     */
    unlock_file(ctx->folder);
    fclose(ctx->folder->fp);
    fclose(src);
    
#ifdef USE_SETGID
    /* Make sure to restore the file back to its orignal group! */
    chown(ctx->folder->name, statbuf.st_uid, statbuf.st_gid);
#endif /* USE_SETGID */
    
    /* Restore the previous access/modification times */
    utimebuf.actime = statbuf.st_atime;
    utimebuf.modtime = statbuf.st_mtime;
    utime(ctx->folder->name, &utimebuf);
  }
  else {
    /* no messages left, so unlink the file */
    unlink(ctx->folder->name);
    unlock_file(ctx->folder);
  }
  ctx->locked = 0;
  unlink(tempfile);

  mutt_unblock_signals();

  Sort = save_sort; /* Restore the default value. */
  mutt_error("%d messages kept, %d deleted.", saved, ctx->msgcount - saved);
  mutt_fastclose_mailbox(ctx);
  return(1);

  /* Come here in case of disaster */
 bail:
  if (ctx->locked) {
    unlock_file(ctx->folder);
    ctx->locked = 0;
  }

  mutt_unblock_signals();

  if ((ctx->folder->fp = freopen(ctx->folder->name, "r", ctx->folder->fp)) == NULL) {
    mutt_error("Could not reopen mailbox!");
    sleep(1);
    mutt_exit(1);
  }
  if (need_sort || save_sort != Sort) {
    Sort = save_sort;
    mutt_sort_headers(ctx);
  }
  dprint(1,(debugfile,"mutt_sync_mailbox(): operation failed.\n"));
  return(0);
}

/*
 * return TRUE if successful, otherwise return FALSE.
 */
int mutt_close_mailbox (CONTEXT *ctx)
{
  int i, modified=0, read=0, deleted=0, move_messages=0, keep_deleted=0;
  FOLDER *f;
  
  if (!ctx->folder) return(1);

  for (i=0; i < ctx->msgcount; i++) {
    if (ctx->hdrs[i]->deleted) {
      deleted++;
      continue;
    }
    else if (! ctx->hdrs[i]->old && option(OPTMARKOLD)) {
      /* Mark new messages as old. */
      ctx->hdrs[i]->old = 1;
      ctx->hdrs[i]->changed = 1;
      modified++;
    }
    else if (ctx->hdrs[i]->changed)
      modified++;
    
    /* Note that deleted messages are not counted. */
    if (ctx->hdrs[i]->read) read++;
  }

  if (read && !option(OPTHOLD) &&
      strcmp(ctx->folder->name, Spoolfile) == 0 &&
      strcmp(Inbox, Spoolfile) != 0) {
    if (option(OPTASKMOVE))
      move_messages = mutt_yesorno("Move read messages to mbox?", 1);
    else
      move_messages = 1;
  }

  if (deleted && option(OPTASKDEL))
    keep_deleted = mutt_yesorno("Keep deleted messages?", 0);

  if (move_messages) {
    if ((f = folder_open(Inbox, "a")) == 0) return(0);

    mutt_block_signals();

    if (lock_file(f, 1) != 0) goto cleanup;

    mutt_error("Moving read messages to %s...", Inbox);

    for (i = 0; i < ctx->msgcount; i++)	{
      if (ctx->hdrs[i]->read && !ctx->hdrs[i]->deleted) {
	if (copy_message(ctx->folder, ctx->hdrs[i], f) == -1) {
	  unlock_file(f);
	  mutt_error("Error while moving messages!");
	  goto cleanup;
	}
	ctx->hdrs[i]->deleted = 1;
      }
    }
    unlock_file(f);
    folder_close(&f);
  }
  else if (!modified && !deleted) {
    mutt_error("Mailbox is unchanged.");
    mutt_fastclose_mailbox(ctx);
    return(1);
  }

  if (keep_deleted) {
    for (i=0; i<ctx->msgcount; i++)
      ctx->hdrs[i]->deleted = 0;
    deleted = 0;
  }

  /*
   * if mutt_sync_mailbox() fails, the user will likely end up with two copies
   * of a message, because we have already appended the "read" messages to
   * the "mbox" mailbox by this point.  however, since nothing will have been
   * lost, this isn't really that big of a problem.
   */
  return(mutt_sync_mailbox(ctx));

cleanup:
  folder_close(&f);
  mutt_unblock_signals();
  return(0);
}

CONTEXT *mutt_open_mailbox (char *fol, CONTEXT *ctx, int flags)
{
  if (ctx == NULL) ctx = (CONTEXT *)safe_malloc(sizeof(CONTEXT));
  memset(ctx, 0, sizeof(CONTEXT));
  ctx->weed = 1;
  mutt_expand_path(fol);
  if ((ctx->folder = folder_open(fol, "r")) == 0) {
    safe_free((void **)&ctx);
    return 0;
  }
  mutt_block_signals();
  if (lock_file(ctx->folder, 0) == -1) {
    folder_close(&ctx->folder);
    mutt_unblock_signals();
    safe_free((void **)&ctx);
    return 0;
  }
  mutt_parse_mailbox(ctx);
  unlock_file(ctx->folder);
  mutt_unblock_signals();

  /* we don't always need to sort the mailbox (like the postponed mailbox) */
  if (flags & M_NOSORT) {
    int i;

    ctx->vcount = 0;
    for (i=0; i<ctx->msgcount; i++) {
      ctx->hdrs[i]->virtual = ctx->vcount;
      ctx->hdrs[i]->msgno = i + 1;
      *ctx->v2r[ctx->vcount] = i;
      ctx->vcount++;
    }
  }
  else
    mutt_sort_headers(ctx);
  return(ctx);
}

/*
 * This routine is used to append the temporary file of an outgoing message
 * the the user's "outbox".
 */
void mutt_folder_cc (const char *fname, const char *folder)
{
  int  len;
  FILE *fin;
  char buffer[LONG_STRING];
  time_t t;
  FOLDER *out;

  if ((fin = fopen(fname, "r")) == NULL) {
    dprint(1, (debugfile, "append_file_to_folder: oops, couldn't open file to be copied!\n"));
    mutt_error("Error opening file to append!");
    sleep(1);
    return;
  }
  if ((out = folder_open(folder, "a")) == 0) {
    dprint(1,(debugfile, "mutt_folder_cc(): unable to open %s\n", folder));
    fclose(fin);
    return;
  }
  mutt_block_signals();
  if (lock_file(out, TRUE) == -1) {
    fclose(fin);
    folder_close(&out);
    mutt_unblock_signals();
    mutt_error("Could not lock mailbox %s!", folder);
    return;
  }
  fseek(out->fp, 0, 2);

  if (out->magic == MAGICMMDF) fputs("\001\001\001\001\n", out->fp);

  t = time(0);
  fprintf (out->fp, "From %s %s", Username, asctime(localtime(&t)));

  /* first copy the header.*/
  while (fgets(buffer, sizeof(buffer), fin) != NULL) {
    if (buffer[0] == '\n')
      break;
    fputs(buffer, out->fp);
  }

  /* add a dummy status field */
  fputs("Status: RO\n", out->fp);
  
  /* add the header terminator */
  fputc('\n', out->fp);

  /* now copy the rest of the message */
  while ((len = fread(buffer, 1, sizeof(buffer), fin)) > 0)
    fwrite(buffer, 1, len, out->fp);

  if (out->magic == MAGICMBOX)
    fputc('\n', out->fp);
  else if (out->magic == MAGICMMDF)
    fputs("\001\001\001\001\n", out->fp);

  fflush(out->fp);
  unlock_file(out);
  folder_close(&out);
  mutt_unblock_signals();
  fclose(fin);
}

/*
** Support nation specific RFCs - pms@romance.kaist.ac.kr
*/
int mutt_save_attachment (BODY *m, char *path)
{
  STATE s;

  mutt_expand_path(path);
  if (access(path, F_OK) == 0) {
    mutt_error("File already exists!");
    return(-1);
  }

  if ((s.fpout = fopen(path, "w")) == NULL) {
    mutt_error("Can't open %s for writing!", path);
    return -1;
  }

  fseek((s.fpin = Context->folder->fp), m->offset, 0);
  s.prefix = 0;

  switch(m->encoding) {
  case ENC8BIT:
  case ENC7BIT:
    if(mutt_decode_own_rfc(&s, m->length, 
       mutt_get_parameter("charset", m->parameter))) break;
  case ENCBINARY:
    mutt_decode_xbit(&s, m->length);
    break;
  case ENCQUOTEDPRINTABLE:
    mutt_decode_quoted(&s, m->length, mutt_is_text_type(m->type, m->subtype));
    break;
  case ENCBASE64:
    mutt_decode_base64(&s, m->length, mutt_is_text_type(m->type, m->subtype));
    break;
  }

  if (fclose(s.fpout) == 0) return 0;
  
  dprint(1,(debugfile, "save_attachment: fclose() returned non-zero! (errno=%d)\n", errno));

  return -1;
}

/*
 * This routine is used to save message(s) to the given folder.  "delete"
 * specifies whether or not successfully saved messages are marked as
 * deleted.
 *
 * If "h" is NULL, this routine will look for selected message.
 */
int mutt_save_message (HEADER *h, char *path, int delete)
{
  FOLDER *fp;
  int rc = 0, i, count = 0;
#ifdef USE_SETGID
  struct stat sb;
  gid_t gid;

  if (stat(path, &sb) == -1) {
    if (errno != ENOENT) {
      dprint(1, (debugfile, "save_to_folder(): stat returned errno %d.\n",
		 errno));
      mutt_error("Could not stat that mailbox!");
      return(-1);
    }
    gid = User_gid;
  }
  else
    gid = sb.st_gid;
#endif /* USE_SETGID */

  if ((fp = folder_open(path, "a")) == NULL) {
    rc = -1;
    goto recover;
  }

  if ((rc = lock_file(fp, TRUE)) == -1) {
    folder_close(&fp);
    goto recover;
  }

  if (h) {
    if ((rc = copy_message(Context->folder, h, fp)) == -1) {
      unlock_file(fp);
      folder_close(&fp);
      goto recover;
    }
    if (delete) {
      h->deleted = 1;
      if (h->tagged) {
	h->tagged = 0;
	Context->tagged = 0;
      }
    }
    count++;
  }
  else if (Context->tagged) {
    for (i = 0; i < Context->msgcount; i++)
      if (Context->hdrs[i]->tagged) {
	if ((rc = copy_message(Context->folder, Context->hdrs[i], fp)) == -1) {
	  unlock_file(fp);
	  folder_close(&fp);
	  goto recover;
	}
	if (delete) {
	  Context->hdrs[i]->deleted = 1;
	  if (Context->hdrs[i]->tagged) {
	    Context->hdrs[i]->tagged = 0;
	    Context->tagged = 0;
	  }
	}
	count++;
      }
  }
  fflush(fp->fp);
  unlock_file(fp);
  folder_close(&fp);

#ifdef USE_SETGID
  chown(path, getuid(), gid);
#endif

 recover:

  if (rc < 0)
    mutt_error("Error while copying message!");
  else
    mutt_error("%s %d message%s to %s.", delete?"Saved":"Copied", count,
	       count > 1 ? "s" : "", path);
  return rc;
}

/*
 * Wipe and remove the requested file.
 */
void mutt_unlink (const char *s)
{
  struct stat stbuf;
  FILE *f;
  char b[LONG_STRING];
  int bytes = 0, chunk;

  if (stat(s, &stbuf) == -1) return;
  if ((f = fopen(s, "r+")) == NULL) return;
  
  while (bytes < stbuf.st_size) {
    chunk = stbuf.st_size - bytes;
    if (chunk > LONG_STRING)
      chunk = LONG_STRING;
    fwrite(b, 1, chunk, f);
    bytes += chunk;
  }
  fclose(f);
  unlink(s);
}
