/*
 * Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include "config.h"
#include "muttlib.h"
#include "rfc822.h"
#include "rfc1522.h"

#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <errno.h>

static char *compatible_charsets[] = {
  "US-ASCII",
  "ISO-8859-1",
  "ISO-8859-2",
  "ISO-8859-3",
  "ISO-8859-4",
  "ISO-8859-5",
  "ISO-8859-6",
  "ISO-8859-7",
  "ISO-8859-8",
  "ISO-8859-9",
  "KOI8-R",
  (char *)0,
};

int mutt_compat_charset (const char *s)
{
  int i;

  for (i=0; compatible_charsets[i]; i++)
    if (strcasecmp(s, compatible_charsets[i]) == 0)
      return(1);
  return(0);
}

BODY *mutt_new_body (void)
{
  BODY *p = (BODY *)safe_malloc(sizeof(BODY));
    
  memset(p, 0, sizeof(BODY));
  p->disposition = DISPATTACH;
  p->use_disp = 1;
  return(p);
}

void mutt_free_body (BODY **p)
{
  BODY *a = *p, *b;

  while (a) {
    b = a;
    a = a->next; 

    if (b->parameter) mutt_free_parameter(&b->parameter);
    if (b->unlink) unlink(b->filename);
    safe_free((void **)&b->filename);
    safe_free((void **)&b->content);
    safe_free((void **)&b->subtype);
    safe_free((void **)&b->description);

    if (b->parts) mutt_free_body(&b->parts);

    safe_free((void **)&b);
  }
  *p = 0;
}

PARAMETER *mutt_new_parameter (void)
{
  PARAMETER *p = (PARAMETER *)safe_malloc(sizeof(PARAMETER));
  
  memset(p, 0, sizeof(PARAMETER));
  return(p);
}

void mutt_free_parameter (PARAMETER **p)
{
  PARAMETER *t = *p;
  PARAMETER *o;

  while (t) {
    safe_free((void **)&t->attribute);
    safe_free((void **)&t->value);
    o = t;
    t = t->next;
    safe_free((void **)&o);
  }
  *p = 0;
}

STATE *mutt_new_state (void)
{
  STATE *p = (STATE *)safe_malloc(sizeof(STATE));

  memset(p, 0, sizeof(STATE));
  return p;
}

void mutt_free_state (STATE **p)
{
  safe_free((void **)p);
}

HEADER *mutt_new_header (void)
{
  HEADER *p = (HEADER *)safe_malloc(sizeof(HEADER));

  memset(p, 0, sizeof(HEADER));
  return(p);
}

void mutt_free_list (LIST **list)
{
  LIST *p;
  
  if (!list) return;
  while (*list) {
    p = *list;
    *list = (*list)->next;
    safe_free((void **)&p->data);
    safe_free((void **)&p);
  }
}

void mutt_free_header (HEADER **h)
{
  mutt_free_envelope(&(*h)->env);
  mutt_free_body(&(*h)->content);
  safe_free((void **)&(*h)->message_id);
  safe_free((void **)&(*h)->tree);
  safe_free((void **)h);
}

void remove_leading_space (char *s)
{
  char *p = s;

  while (*p && isspace(*p)) p++;
  if (p != s && *p) {
    while (*p) *s++ = *p++;
  }
}

/* returns true if the given header field is named in the user's ignore list */
int mutt_matches_ignore (const char *s)
{
  LIST *t = Ignore;

  while (t) {
    if (strncasecmp(s, t->data, strlen(t->data)) == 0)
      return 1;
    t = t->next;
  }
  return 0;
}

void mutt_expand_path (char *s)
{
  char buf[_POSIX_PATH_MAX];

  if (*s == '~')
    sprintf(buf, "%s%s", Homedir, s+1);
  else if (*s == '=' || *s == '+')
    sprintf(buf, "%s/%s", Maildir, s+1);
  else if (strcmp(s, ">") == 0)
    strcpy(buf, Inbox);
  else if (strcmp(s, "<") == 0)
    strcpy(buf, Outbox);
  else if (strcmp(s, "!") == 0)
    strcpy(buf, Spoolfile);
  else
    return;
  strcpy(s, buf); /* replace the string with the expanded version. */
}

/* returns TRUE if the given address belongs to the user. */
int mutt_addr_is_user (ADDRESS *addr)
{
  ADDRESS *a = Alternates;
  int l;

  /* NULL address is assumed to be the user. */
  if (!addr) return 1;

  if (!addr->mailbox) return 0;
  if (!addr->host || *addr->host == '@') {
    if (strcasecmp(addr->mailbox, Username) == 0) return 1;
    return 0;
  }

  if (strcasecmp(addr->mailbox, Username) == 0 &&
      (strcasecmp(addr->host, Hostname) == 0 ||
       strcasecmp(addr->host, Fqdn) == 0))
    return 1;

  l = strlen(addr->host);

  while (a) {
    if (strcasecmp(addr->mailbox, a->mailbox) == 0) {
      /*
       * From: David DeSimone <fox@mailhost.rsn.hp.com>
       * Check for pseudo-pattern matching.  A user can have an alternate
       * address like ``me@*.cs.hmc.edu'' which should match any host in
       * the cs.hmc.edu domain.
       */
      if (a->host[0] == '*') {
	int r = strlen(&a->host[1]);
	
	if ((l >= r) && (strcasecmp(&addr->host[l-r], &a->host[1]) == 0))
	  return 1;
      }
      else if (strcasecmp(addr->host, a->host) == 0)
	return 1;
    }
    a = a->next;
  }
  return(0);
}

void *safe_malloc (unsigned int siz)
{
  void *p;

  if (siz == 0) return 0;
  if ((p = (void *)malloc(siz)) == 0) {
    mutt_error("Out of memory!");
    sleep(1);
    mutt_exit(1);
  }
  return(p);
}

void safe_realloc (void **p, unsigned int siz)
{
  void *r;

  if (siz == 0) {
    safe_free((void **)&*p);
    return;
  }
  if ((r = (void *)realloc(*p, siz)) == 0) {
    mutt_error("Out of memory!");
    sleep(1);
    mutt_exit(1);
  }
  *p = r;
}

void safe_free (void **p)
{
  if (!*p) return;
  free(*p);
  *p = 0;
}

char *safe_strdup (const char *s)
{
  char *p;

  /* Do not return for NUL string - pms@romance.kaist.ac.kr */
  if (!s) return 0;
  p = (char *)safe_malloc(strlen(s)+1);
  strcpy(p, s);
  return(p);
}

char *skip_whitespace (char *p)
{
  while (*p && (*p == ' ' || *p == '\t')) p++;
  return(p);
}

void mutt_copy_bytes (FILE *in, FILE *out, long size)
{
  char buffer[LONG_STRING];
  long chunk;

  while (size > 0) {
    if (size > sizeof(buffer))
      chunk = sizeof(buffer);
    else
      chunk = size;
    if ((chunk = fread(buffer, 1, chunk, in)) < 1) break;
    fwrite(buffer, 1, chunk, out);
    size -= chunk;
  }
}

ADDRESS *mutt_new_address (void)
{
  ADDRESS *p = (ADDRESS *)safe_malloc(sizeof(ADDRESS));
  
  memset(p, 0, sizeof(ADDRESS));
  return(p);
}

void mutt_free_address (ADDRESS **p)
{
  ADDRESS *t;
  ADDRESS *o;

  if (!*p) return;
  t = *p;
  while (t) {
    safe_free((void **)&t->personal);
    safe_free((void **)&t->mailbox);
    safe_free((void **)&t->adl);
    safe_free((void **)&t->host);
    safe_free((void **)&t->error);
    o = t;
    t = t->next;
    safe_free((void **)&o);
  }
  *p = 0;
}

char *mutt_get_parameter (const char *s, PARAMETER *p)
{
  for (; p; p=p->next)
    if (strcasecmp(s, p->attribute) == 0)
      return(p->value);
  return(0);
}

#ifdef METAMAIL
int mutt_needs_metamail (BODY *m)
{
  switch (m->type) {
  case TYPETEXT:
  case TYPEMULTIPART:
  case TYPEMESSAGE:
    return 0;
  default:
    return 1;
  }
}
#endif /* METAMAIL */

int mutt_is_text_type (int t, char *s)
{
  if (t == TYPETEXT)
    return 1;
  if (t == TYPEMESSAGE) {
    if (strcasecmp("delivery-status", s) == 0)
      return 1;
  }
  return 0;
}

void mutt_default_savename (char *buf, size_t buflen, ENVELOPE *cur)
{
  ADDRESS *a = cur->to;

  *buf++ = '=';
  buflen--;
  
  /* First check to see whether or not this message is addressed to one of
   * the user's known mailing lists! */
  while (a) {
    if (mutt_is_mail_list(a)) {
      strfcpy(buf, a->mailbox, buflen);
      goto do_lower;
    }
    a = a->next;
  }

  a = cur->cc;
  while (a) {
    if (mutt_is_mail_list(a)) {
      strfcpy(buf, a->mailbox, buflen);
      goto do_lower;
    }
    a = a->next;
  }

  if (cur->reply_to)
    strfcpy(buf, cur->reply_to->mailbox, buflen);
  else if (cur->from)
    strfcpy(buf, cur->from->mailbox, buflen);
  else if (cur->to)
    strfcpy(buf, cur->to->mailbox, buflen);
  else if (cur->cc)
    strfcpy(buf, cur->cc->mailbox, buflen);
  else {
    /* this should never happen, but... */
    *(buf-1) = 0;
    return;
  }
do_lower:

  mutt_strlower(buf);
}

ENVELOPE *mutt_new_envelope (void)
{
  ENVELOPE *p = (ENVELOPE *)safe_malloc(sizeof(ENVELOPE));

  memset(p, 0, sizeof(ENVELOPE));
  return(p);
}

void mutt_free_envelope (ENVELOPE **p)
{
  if (!*p) return;
  mutt_free_address(&(*p)->to);
  mutt_free_address(&(*p)->cc);
  mutt_free_address(&(*p)->bcc);
  mutt_free_address(&(*p)->sender);
  mutt_free_address(&(*p)->from);
  mutt_free_address(&(*p)->reply_to);
  safe_free((void **)&(*p)->subject);
  mutt_free_list(&(*p)->references);
  mutt_free_list(&(*p)->userhdrs);
  safe_free((void **)p);
}

char *mutt_make_date (char *s)
{
  time_t t = time(0);
  struct tm *l = gmtime(&t);
  int yday = l->tm_yday;
  int tz = l->tm_hour * 60 + l->tm_min;

  l = localtime(&t);
  tz = l->tm_hour * 60 + l->tm_min - tz;
  yday = l->tm_yday - yday;

  if (yday != 0)
    tz += yday * 24 * 60; /* GMT is next or previous day! */

  sprintf(s, "Date: %s, %d %s %d %02d:%02d:%02d %+03d%02d\n",
    Weekdays[l->tm_wday], l->tm_mday, Months[l->tm_mon], l->tm_year+1900,
    l->tm_hour, l->tm_min, l->tm_sec, tz/60, abs(tz) % 60);
  return s;
}

void mutt_pipe_message (HEADER *h, const char *cmd)
{
  int status;
  FILE *f;

  mutt_create_filter(cmd, &f, NULL);
  fseek(Context->folder->fp, h->offset, 0);
  mutt_copy_bytes(Context->folder->fp, f, h->length);
  fclose(f);
  wait(&status);
}

void mutt_tabs_to_spaces (char *s)
{
  while (*s) {
    if (*s == '\t')
      *s=' ';
    s++;
  }
}

void mutt_mktemp (char *s)
{
  sprintf(s, "%s/muttXXXXXX", Tempdir);
  mktemp(s);
}

/* convert all characters in the string to lowercase */
char *mutt_strlower (char *s)
{
  char *p = s;

  while (*p) {
    *p = tolower(*p);
    p++;
  }
  return(s);
}

/* a wrapper around rfc822_write_address() to prevent buffer overrun */
void mutt_write_address (char *buf, size_t buflen, ADDRESS *addr)
{
  char tmp[LONG_STRING]; /* should be long enough to hold a single address */
  ADDRESS *tmpaddr;
  int len;

  buf[buflen-1] = 0;
  buflen--;

  /* move past what is already in the buf */
  if (buf[0]) {
    len = strlen(buf);
    buflen -= len;
    buf += len;

    if (buflen < 2) return; /* not enough room! */
    *buf++ = ',';
    *buf++ = ' ';
    buflen -= 2;
  }

  while (addr) {
    tmpaddr = addr->next;
    addr->next = NULL;
    tmp[0] = 0;
    rfc822_write_address(tmp, addr);
    if (tmpaddr) strcat(tmp, ", ");
    addr->next = tmpaddr;
    addr = addr->next;
    len = strlen(tmp);
    if (len < buflen) {
      strcpy(buf, tmp);
      buflen -= len;
      buf += len;
    }
    else
      return; /* no more room */
  }
}

/* a wrapper around rfc822_parse_adrlist to check for errors */
int mutt_parse_adrlist (ADDRESS **a, char *s, char *host)
{
  ADDRESS *p;
  ADDRESS *t = 0;
  char tmp[LONG_STRING];

  Parse_error[0] = 0;
  strfcpy(tmp, s, sizeof(tmp)); /* make a copy of the list */
  rfc822_parse_adrlist(&t, tmp, host);
  if (Parse_error[0]) {
    mutt_error(Parse_error);
    mutt_free_address(&t);
    return(-1);
  }
  else {
    /* append the new addresses to the list */
    if (*a) {
      p = *a;
      while (p->next) p = p->next;
      p->next = t;
    }
    else
      *a = t;
    return 0;
  }
  /* not reached */
}

int mutt_is_valid_mailbox (const char *path)
{
  FILE *fp;
  char buf[SHORT_STRING];

  if (access(path, R_OK|F_OK) != 0) {
    if (errno == EACCES)
      mutt_error("Permission denied.");
    else
      mutt_error("%s does not exist.", path);
    return(0);
  }
  if ((fp = fopen(path, "r")) == NULL) return(0);
  if (fgets(buf, sizeof(buf), fp) == NULL) {
    /* empty mailbox, trust the user that this is ok */
    return(1);
  }
  if (strncmp("From ", buf, 5) == 0) return(1);
  mutt_error("%s does not appear to be a mailbox.", path);
  return(0);
}

/* write an address without the personal name */
void mutt_simple_address (char *buf, size_t buflen, ADDRESS *addr)
{
  int len;
  
  *buf=0;
  if (!addr) return;
  strfcpy(buf, addr->mailbox, buflen);
  len = strlen(buf);
  buflen -= len;
  buf += len;
  if (*addr->host != '@') snprintf(buf, buflen, "@%s", addr->host);
}

int option (int n)
{
  if (n < 32)
    return(OptionsA & (1 << n));
  else
    return(OptionsB & (1 << (n - 31)));
}

LIST *mutt_new_list (void)
{
  LIST *p = safe_malloc(sizeof(LIST));

  memset(p, 0, sizeof(LIST));
  return p;
}

int is_local_site (const char *s)
{
  LIST *tmp = Localsites;
  const char *a;
  char *b;
  
  if (!s || *s == '@' || strcasecmp(s, Fqdn) == 0) return 1;
  while (tmp) {
    if (*tmp->data == '*') {
      a = strchr(s, '.');
      b = tmp->data + 1;
    }
    else {
      a = s;
      b = tmp->data;
    }
    if (a && b && strcasecmp(a, b) == 0) return 1;
    tmp = tmp->next;
  }
  return 0;
}
