/*
 * Copyright (C) 1996 Michael R. Elkins <me@cs.hmc.edu>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "mutt.h"
#include "mutt_curses.h"
#include "mutt_menu.h"

#include <string.h>

#define OFFSET 1
#define LENGTH (LINES-4)

MUTTMENU *mutt_newMenu (const char *title)
{
  MUTTMENU *p = (MUTTMENU *)safe_malloc(sizeof(MUTTMENU));

  memset(p, 0, sizeof(MUTTMENU));
  p->title = safe_strdup(title);
  p->redraw = REDRAW_FULL;
#ifdef NCURSES_MOUSE_VERSION
  if (Mousemask) toolbar_init(Generic_toolbar);
#endif /* NCURSES_MOUSE_VERSION */
  return(p);
}

void mutt_menuDestroy (MUTTMENU **p)
{
  safe_free((void **)&(*p)->title);
  safe_free((void **)p);
}

int mutt_menuLoop (MUTTMENU *menu)
{
  int i;
  char buffer[STRING];

  FOREVER {
    /*
     * See if all or part of the screen needs to be updated.
     */
    switch (menu->redraw) {
    case REDRAW_FULL:
      clear();
#ifdef NCURSES_MOUSE_VERSION
      if (Mousemask) toolbar_post(Generic_toolbar);
#endif
      SETCOLOR(MT_COLOR_STATUS);
      mvprintw(LINES-2, 0, "%-*.*s", COLS, COLS, "");
      mvaddstr(LINES-2, 0, menu->title);
      SETCOLOR(MT_COLOR_NORMAL);

      /* fall through to REDRAW_INDEX */

    case REDRAW_INDEX:
      for (i = menu->top; i < menu->top + LENGTH; i++) {
	if (i < menu->max) {
	  menu->makeEntry(buffer, i, sizeof(buffer));
	  buffer[COLS-3] = 0;
	  CLEARLINE(i - menu->top + OFFSET);
          move(i - menu->top + OFFSET, 3);
          addstr(buffer);
	}
	else
	  CLEARLINE(i-menu->top+OFFSET);
      }
      menu->redraw = 0;
      break;

    case REDRAW_MOTION:
      move(menu->oldcurrent + OFFSET - menu->top, 0);
      addstr("  ");
      menu->redraw = 0;
      break;
    }

    SETCOLOR(MT_COLOR_INDICATOR);
    move(menu->current - menu->top + OFFSET, 0);
    addstr("->");
    SETCOLOR(MT_COLOR_NORMAL);

    refresh();
    i = ci_getch();
    ci_clear_error();

#ifdef NCURSES_MOUSE_VERSION
    if (i == KEY_MOUSE) {
      MEVENT m;

      getmouse(&m);
      if (m.y == 0) {
	if ((i = toolbar_select(Generic_toolbar, m.x)) == -1) continue;
      }
      else if (m.y >= OFFSET && m.y <= OFFSET + LENGTH) {
	menu->oldcurrent = menu->current;
	menu->current = m.y - OFFSET + menu->top;
	menu->redraw = REDRAW_MOTION;
	if (m.bstate & BUTTON1_DOUBLE_CLICKED)
	  return(OP_GENERIC_SELECT_ENTRY);
	continue;
      }
      else
	continue;
    }
#endif /* NCURSES_MOUSE_VERSION */

    i = keymap_lookup(i, menu->menu);

    switch (i) {
    case OP_GENERIC_NEXT_ENTRY:
      if (menu->current < menu->max - 1) {
	menu->oldcurrent = menu->current;
	menu->current++;
	if (menu->current >= menu->top + LENGTH) {
	  menu->top = menu->current;
	  menu->redraw = REDRAW_INDEX;
	}
	else
	  menu->redraw = REDRAW_MOTION;
      }
      else {
	mutt_error("You are on the last entry.");
      }
      break;

    case OP_GENERIC_PREVIOUS_ENTRY:
      if (menu->current > 0) {
	menu->oldcurrent = menu->current;
	if (--menu->current < menu->top) {
	  if ((menu->top = menu->current - LENGTH + 1) < 0) menu->top = 0;
	  menu->redraw = REDRAW_INDEX;
	}
	else
	  menu->redraw = REDRAW_MOTION;
      }
      else {
	mutt_error("You are on the first entry.");
      }
      break;
	
    case OP_GENERIC_NEXT_PAGE:
      if (menu->top + LENGTH <= menu->max - 1) {
	menu->top += LENGTH;
	menu->current = menu->top;
	menu->redraw = REDRAW_INDEX;
      }
      else
	mutt_error("You are on the last page.");
      break;

    case OP_GENERIC_PREVIOUS_PAGE:
      if (menu->top > 0) {
	if ((menu->top -= LENGTH) < 0) menu->top = 0;
	menu->current = menu->top + LENGTH - 1;
	menu->redraw = REDRAW_INDEX;
      }
      else
	mutt_error("You are on the first page.");
      break;

    case OP_GENERIC_FIRST_ENTRY:
      menu->oldcurrent = menu->current;
      if (menu->top > 0)
	menu->redraw = REDRAW_INDEX;
      else
	menu->redraw = REDRAW_MOTION;
      menu->top = menu->current = 0;
      break;

    case OP_GENERIC_LAST_ENTRY:
      menu->oldcurrent = menu->current;
      menu->current = menu->max - 1;
      if (menu->current > menu->top + LENGTH - 1) {
	menu->top = menu->current - LENGTH;
	if (menu->top < 0) menu->top = 0;
	menu->redraw = REDRAW_INDEX;
      }
      else
	menu->redraw = REDRAW_MOTION;
      break;

    case OP_GENERIC_TOP_OF_PAGE:
      menu->oldcurrent = menu->current;
      menu->current = menu->top;
      menu->redraw = REDRAW_MOTION;
      break;

    case OP_GENERIC_MIDDLE_OF_PAGE:
      menu->oldcurrent = menu->current;
      i = menu->top + LENGTH;
      if (i > menu->max - 1)
	i = menu->max - 1;
      menu->current = menu->top + (i - menu->top) / 2;
      break;

    case OP_GENERIC_BOTTOM_OF_PAGE:
      menu->oldcurrent = menu->current;
      menu->current = menu->top + LENGTH - 1;
      if (menu->current > menu->max - 1)
	menu->current = menu->max - 1;
      menu->redraw = REDRAW_MOTION;
      break;

    case OP_GENERIC_REFRESH:
      menu->redraw = REDRAW_FULL;
      break;

    case OP_GENERIC_HELP:
      ci_help(menu->menu);
      menu->redraw = REDRAW_FULL;
      break;

    case OP_NULL:
      mutt_error("Key is not bound.  Press '?' for help.");
      break;

    default:
      return (i);
    }
  }
  /* not reached */
}
