/*
 * Copyright (C) 1996 Michael R. Elkins <me@cs.hmc.edu>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 


#include "mutt.h"
#include "mutt_curses.h"
#include "keymap.h"

#include <sys/stat.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#define MUTT_TABSTOP 8
#define next_tabstop (col > (COLS/8)*8 ? 0 : (((col/MUTT_TABSTOP)+1)*MUTT_TABSTOP))

#define ISHEADER(x) ((x == MT_COLOR_HEADER) || (x == MT_COLOR_SUBJECT) || (x == MT_COLOR_FROM))

struct line_t {
  long  offset; /* position in the file */
  short length; /* how many characters on this line? */
  short class;  /* what type is it? */
};

#define pclear move(1,0),clrtobot
#define setpos(x) fseek(fp, offset[x]->offset, 0);

/* calclulate the length of a line after expanding tabs and control chars */
static int calc_length (char *s)
{
  int len = 0;
  int col = 0;
  int nextcol;

  while (*s) {
    if (col == COLS) col = 0;
    if (*s == '\t') {
      nextcol = next_tabstop;
      if (nextcol == 0) {
	len += COLS - col;
	col = 0;
      }
      else {
	for (; col<nextcol; col++) 
	  len++;
      }
    }
    else if (*s == '\n')
      break; /* end of line */
    else if (iscntrl(*s)) {
      if (col > COLS - 1) col = COLS - col;
      len += 2;
      s++;
    }
    else {
      col++;
      len++;
    }
    s++;
  }
  return(len);
}

/*
 * returns the line that is ``n'' lines before ``cur'', taking wrapped lines
 * and hidden text into consideration
 */
static int upNLines (int n, struct line_t **p, int cur, int hiding)
{
  while (cur > 0 && n > 0) {
    cur--;
    if (!hiding || (hiding && p[cur]->class != MT_COLOR_QUOTED)) {
      n -= p[cur]->length / COLS;
      if (!p[cur]->length || p[cur]->length % COLS != 0) n--;
    }
  }
  return(cur);
}

int ci_simple_pager (const char *banner, const char *fname, int do_color)
{
  unsigned char buffer[LONG_STRING];
  unsigned char *pending=0;
  int i, ch = 0, rc = -1, col = 0;
  int last_match = -1; /* last line that matched the search pattern */
  int lines=0, curline=0;
  FILE *fp=NULL;
  struct stat sb;
  struct line_t **offset = NULL; /* array of lines */
  int offsetLen = 0; /* number of used entries in offset */
  int offsetMax = 0; /* number of entries in offset */
  int attr;
  int hideQuoted = 0; /* reap quoted text? */
  int redraw = 1;
  int err;
  int topline = 0;
#ifdef HAVE_REGCOMP
  regex_t quote_re; /* how to find quoted lines in the body of a message */
  regex_t re;
#elif HAVE_REGCMP
  char *quote_re;
  char *re;
#endif
  static char searchbuf[STRING];
  int searching=0, savecurline=0;

  if ((fp = fopen(fname, "r")) == NULL) {
    mutt_error("Could not open file %s!", fname);
    return(-1);
  }
  if (stat(fname, &sb) != 0) {
    fclose(fp);
    return(-1);
  }
  unlink(fname);

  /* compile the expression for finding quoted lines */
  if ((err = REGCOMP(quote_re, QuoteString)) != 0) {
#ifdef HAVE_REGCOMP
    regerror(err, &quote_re, buffer, sizeof(buffer));
    mutt_error(buffer);
#else
    mutt_error("quote_regexp contains an invalid expression.");
#endif
    REGFREE(quote_re);
    fclose(fp);
    return(-1);
  }

  while (ch != -1) {
    if (redraw) {
      clear();

      /* print the banner on the top of the screen */
      SETCOLOR(MT_COLOR_STATUS);
      move(0,0);
      printw("%-*.*s", COLS, COLS, banner);
      SETCOLOR(MT_COLOR_NORMAL);

      curline = topline;
      lines = 0;
      if (offsetLen) {
        setpos(curline);
      }
      move(1, 0);

      redraw = 0;
    }

    while (lines < LINES - 2) {

      /* allocate more space if needed */
      if (offsetMax == offsetLen) {
	if (offset == NULL)
	  offset = safe_malloc(sizeof(struct line_t *)*((offsetMax = LINES)));
	else {
	  offsetMax += LINES;
	  safe_realloc((void **)&offset, sizeof(struct line_t *) * offsetMax);
	}
      }
      
      if (curline == offsetLen) {
	offset[curline] = safe_malloc(sizeof(struct line_t));
	offset[curline]->offset = ftell(fp);
	offset[curline]->class = -1;
	offsetLen++;
      }

      if (fgets(buffer, sizeof(buffer), fp) == NULL) {
	if (lines == -1) {
	  pclear();
	  lines=0;
	  topline = curline = upNLines(LINES-2, offset, curline, hideQuoted);
	  setpos(curline);
	  continue;
	}
	break;
      }

      /* Attempt to figure out what "class" this line is. */
      if (offset[curline]->class == -1) {
	if (curline == 0 ||
	    (curline > 0 && ISHEADER(offset[curline-1]->class) && buffer[0] != '\n')) {
	  if (strncasecmp("subject:", buffer, 8) == 0)
	    offset[curline]->class = MT_COLOR_SUBJECT;
	  else if (strncasecmp("from:", buffer, 5) == 0)
	    offset[curline]->class = MT_COLOR_FROM;
	  else
	    offset[curline]->class = MT_COLOR_HEADER;
	}
	else if (strncmp(buffer, "[Attachment #", 13) == 0)
	  offset[curline]->class = MT_COLOR_ATTACHMENT;
	else if (strcmp("-- \n", buffer) == 0)
	  offset[curline]->class = MT_COLOR_SIGNATURE;
	else if (curline > 0 && offset[curline-1]->class == MT_COLOR_SIGNATURE)
	  offset[curline]->class = MT_COLOR_SIGNATURE;
	else if (REGEXEC(quote_re, buffer) == 0)
	  offset[curline]->class = MT_COLOR_QUOTED;
	else
	  offset[curline]->class = MT_COLOR_NORMAL;
	
	/* keep track of how long this line is */
	offset[curline]->length = calc_length(buffer);
      }

      if (searching) {
	if (REGEXEC(re, buffer) == 0) {
	  pclear();
	  searching = 0;
	  last_match = topline = curline;
	  setpos(curline);
#ifdef HAVE_REGCOMP
	  regfree(&re);
#endif
	  continue;
	}
	curline++;
	continue;
      }
      /*
       * lines == -1 has the special meaning that we are looking for the end
       * of the file
       */
      if (lines == -1) {
	curline++;
	continue;
      }

      /* skip quoted text */
      if (hideQuoted && offset[curline]->class == MT_COLOR_QUOTED) {
	col = 0;
	curline++;
	continue;
      }

      if (lines + offset[curline]->length / COLS + (offset[curline]->length % COLS != 0 ? 1 : 0) > LINES - 1) {
	/* not enough room to display the whole line */
	break;
      }

      pending = buffer;
      attr = offset[curline]->class;
      if (do_color) SETCOLOR(attr);
      col = 0;
      while (*pending) {
	if (col == COLS) {
	  lines++;
	  col = 0;
	  /* If the last character on the line is right at the edge of
	   * the screen, don't print the newline, or else we'll end up
	   *  with a blank line.
	   */
	  if (*pending == '\n') pending++;
	  continue;
	}
	if (*pending == '\n') {
	  addch(*pending);
	  col = 0;
	  lines++;
	}
	else if (*pending == '\t') {
	  i = next_tabstop;
	  if (i == 0) {
	    addch('\n');
	    lines++;
	    col = 0;
	  }
	  else
	    for (i=i-col; i > 0; i--, col++) addch(' ');
	}
	else if (IsPrint(*pending)) {
	  addch(*pending);
	  col++;
	}
	else if (iscntrl(*pending)) {
	  if (col > COLS-1) {
	    addch('\n');
	    col = 0;
	    lines++;
	    continue;
	  }
	  addch('^');
	  addch(*pending+'@');
	  col += 2;
	  if (*pending == ctrl('L')) {
	    pending++;
	    break;
	  }
	}
	else
	  addch('?'); /* unprintable character! */
	pending++;
      }

      /* finished with this line */
      curline++;
    }

    if (do_color) SETCOLOR(MT_COLOR_NORMAL);

    if (searching) {
      CLEARLINE(LINES-1);
      attrset(A_REVERSE);
      addstr("Not found!");
      SETCOLOR(MT_COLOR_NORMAL);
      searching=0;
      curline=savecurline;
      setpos(curline);
      REGFREE(re);
    }
    else {
      move(LINES-1, 0);
      clrtoeol();
      SETCOLOR(MT_COLOR_STATUS);
      if (ftell(fp) < sb.st_size-1)
	printw("More (you've seen %d%%)", 100*ftell(fp)/sb.st_size);
      else
	addstr("(END)");
      SETCOLOR(MT_COLOR_NORMAL);
    }
    
    refresh();
    ch = ci_getch();

#ifdef NCURSES_MOUSE_VERSION
    if (ch == KEY_MOUSE) {
      MEVENT e;
      
      getmouse(&e);
      if (e.bstate & BUTTON1_CLICKED) {
	if (e.y <= LINES/2)
	  ch = OP_PAGER_NEXT_PAGE;
	else
	  ch = OP_PAGER_PREVIOUS_PAGE;
      }
      else
	continue;
      }
    else
#endif
      ch = keymap_lookup(ch, MENU_PAGER);

    rc = ch;

    switch (ch) {
    case OP_PAGER_EXIT:

      rc = -1;
      ch = -1;
      break;

    case OP_PAGER_NEXT_PAGE:

      topline = curline;
      if (ftell(fp) < sb.st_size-1) {
	lines = 0;
	pclear();
      }
      else if (option(OPTPAGERSTOP)) {
	/* emulate "less -q" and don't go on to the next message. */
	beep();
      }
      else {
	/* end of the current message, so display the next message. */
	rc = OP_MAIN_NEXT_UNDELETED;
	ch = -1;
      }
      break;

    case OP_PAGER_PREVIOUS_PAGE:

      if (topline != 0) {
	pclear();
	topline = curline = upNLines(LINES-2, offset, topline, hideQuoted);
	setpos(curline);
	lines = 0;
      }
      break;

    case OP_PAGER_NEXT_LINE:

      if (ftell(fp) < sb.st_size-1) {
	pclear();
	topline++;
	curline = topline;
	setpos(curline);
	lines = 0;
      }
      else
	beep();
      break;

    case OP_PAGER_PREVIOUS_LINE:

      pclear();
      if (topline) topline--;
      curline = topline;
      setpos(curline);
      lines = 0;
      break;

    case OP_PAGER_TOP:

      pclear();
      lines=0;
      fseek(fp, 0, 0);
      topline = curline = 0;
      break;

    case OP_PAGER_SEARCH:

      /* modified - pms@romance.kaist.ac.kr */
      if (ci_get_field("Search: ", searchbuf, sizeof(searchbuf), CI_newsearch) != 0 || !searchbuf[0]) break;

      if ((err = REGCOMP(re, searchbuf)) != 0) {
#ifdef HAVE_REGCOMP
	char error[STRING];
	
	regerror(err, &re, error, sizeof(error));
	mutt_error(error);
#else
	mutt_error("Error in regular expression.");
#endif
	REGFREE(re);
	searching=0;
      }
      else {
	if (topline < offsetLen - 1) {
	  savecurline = curline;
	  curline = topline + 1;
	  setpos(curline);
	  searching = 1;
	  lines=0;
	}
	else {
	  beep();
	  mutt_error("You are at the end of the message!");
	  sleep(1);
	}
      }
      break;

    case OP_PAGER_HELP:

      ci_help(MENU_PAGER);
      redraw = 1;
      break;

    case OP_PAGER_HIDE_QUOTED:
      
      hideQuoted = !hideQuoted;
      topline = curline = upNLines(LINES-2, offset, curline, hideQuoted);
      setpos(curline);
      lines=0;
      pclear();
      break;

    case OP_PAGER_BOTTOM: /* move to the end of the file */

      if (ftell(fp) < sb.st_size - 1)
	lines = -1;
      else
	beep();
      break;

    default:
      ch = -1;
    }
  }
  fclose(fp);
  for (i=0; i<offsetLen; i++) safe_free((void **)&offset[i]);
  safe_free((void **)&offset);
  REGFREE(quote_re);
  return (rc != -1 ? rc : 0);
}
