/*
 * Copyright (C) 1996 Michael R. Elkins <me@cs.hmc.edu>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include "muttlib.h"
#include "rfc822.h"
#include "rfc1522.h"

#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <stdlib.h>

void mutt_parse_references (LIST **l, char *s)
{
  LIST *t, *r = *l;

  if (r) while (r->next) r = r->next;
  while ((s = strtok(s, " \t")) != 0) {
    /*
     * some mail clients add other garbage besides message-ids, so do a quick
     * check to make sure this looks like a valid message-id
     */
    if (*s == '<') {
      t = (LIST *)safe_malloc(sizeof(LIST));
      t->data = safe_strdup(s);
      t->next = NULL;
      if (*l) {
	r->next = t;
	r = r->next;
      }
      else
	*l = r = t;
    }
    s = NULL;
  }
}

static int check_encoding (const char *c)
{
  if (strncasecmp("7bit", c, sizeof("7bit")-1) == 0)
    return(ENC7BIT);
  else if (strncasecmp("8bit", c, sizeof("8bit")-1) == 0)
    return(ENC8BIT);
  else if (strncasecmp("binary", c, sizeof("binary")-1) == 0)
    return(ENCBINARY);
  else if (strncasecmp("quoted-printable", c, sizeof("quoted-printable")-1) == 0)
    return(ENCQUOTEDPRINTABLE);
  else if (strncasecmp("base64", c, sizeof("base64")-1) == 0)
    return(ENCBASE64);
  else
    return(ENCOTHER);
}

static PARAMETER *parse_parameters (char *s)
{
  PARAMETER *head = 0, *cur = 0, *new;
  char buffer[STRING], *p, savechar;
  short i;

  while (*s) {
    if ((p = strchr(s, '=')) == 0) {
      dprint(1, (debugfile, "parse_parameters: malformed parameter: %s\n", s));
      return(head); /* just bail out now */
    }
    savechar = *p;
    *p = 0;
    new = mutt_new_parameter();
    new->attribute = safe_strdup(s);
    *p = savechar;
    p = skip_whitespace(p+1);
    if (*p == '"') {
      p++;
      for (i=0; *p && *p != '"'; i++, p++) {
        if (*p == '\\') {
          /* Quote the next character */
          buffer[i] = *(p+1);
          p++;
	  if (!*p) break;
        }
        else
          buffer[i] = *p;
      }
      buffer[i] = 0;
    }
    else {
      for (i=0; *p && *p != ' ' && *p != ';'; i++, p++) {
        buffer[i] = *p;
      }
      buffer[i] = 0;
    }
    new->value = safe_strdup(buffer);

    /* Add this parameter to the list */
    if (head) {
      cur->next = new;
      cur = cur->next;
    }
    else
      head = cur = new;

    /* Find the next parameter */
    if (*p == ';')
      p++;
    else {
      while (*p && *p != ';') p++;
      if (!*p) break; /* no more parameters! */
      p++;
    }

    /* Move past any leading whitespace */
    s = skip_whitespace(p);
  }    
  return(head);
}

int mutt_check_mime_type (const char *s)
{
  if (strcasecmp("text", s) == 0)
    return TYPETEXT;
  else if (strcasecmp("image", s) == 0)
    return TYPEIMAGE;
  else if (strcasecmp("audio", s) == 0)
    return TYPEAUDIO;
  else if (strcasecmp("video", s) == 0)
    return TYPEVIDEO;
  else if (strcasecmp("multipart", s) == 0)
    return TYPEMULTIPART;
  else if (strcasecmp("application", s) == 0)
    return TYPEAPPLICATION;
  else if (strcasecmp("message", s) == 0)
    return TYPEMESSAGE;
  else
    return TYPEOTHER;
}

static void parse_content_type (char *s, BODY *ct)
{
  char *pc;
  char buffer[SHORT_STRING];
  short i = 0;

  safe_free((void **)&ct->subtype);
  mutt_free_parameter(&ct->parameter);

  /* First extract any existing parameters */
  if ((pc = strchr(s, ';')) != 0) {
    *pc = 0;
    pc++;
    while (*pc && whitespace(*pc)) pc++;
    ct->parameter = parse_parameters(pc);

    /* Some pre-RFC1521 gateways still use the "name=filename" convention */
    if ((pc = mutt_get_parameter("name", ct->parameter)) != 0)
      ct->filename = safe_strdup(pc);
  }
  
  /* Now get the subtype */
  pc = strchr(s, '/');
  if (pc) {
    *pc = 0;
    pc++;
    while (*pc && !whitespace(*pc) && *pc != ';') {
      buffer[i++] = *pc;
      pc++;
    }
    buffer[i] = 0;
    ct->subtype = safe_strdup(buffer);
  }

  /* Finally, get the major type */
  ct->type = mutt_check_mime_type(s);

  if (ct->subtype == NULL) {
    /* Some older non-BODY mailers (i.e., mailtool, elm) have a content-type
     * field, so we can attempt to convert the type to BODY here.
     */
    if (ct->type == TYPETEXT)
      ct->subtype = safe_strdup("plain");
    else if (ct->type == TYPEAUDIO)
      ct->subtype = safe_strdup("basic");
    else if (ct->type == TYPEMESSAGE)
      ct->subtype = safe_strdup("rfc822");
    else if (ct->type == TYPEOTHER) {
      ct->type = TYPEAPPLICATION;
      sprintf(buffer, "x-%s", s);
      ct->subtype = safe_strdup(buffer);
    }
  }
}

BODY *mutt_read_mime_header (FILE *fp)
{
  BODY *p = mutt_new_body();
  short linelen = 0;
  char *c;
  char buffer[LONG_STRING], line[HUGE_STRING];
  PARAMETER *parms;
  
  line[0] = 0;
  p->hdr_offset = ftell(fp);
  p->encoding = ENC7BIT; /* default from RFC1521 */
  p->type = TYPETEXT;
  while (fgets(buffer, LONG_STRING, fp) != NULL) {

    if (whitespace(buffer[0]) && buffer[0] != '\n' &&
        ! (buffer[0] == '\r' && buffer[1] == '\n')) {
      /* Remove the trailing whitespace from line */
      c = buffer + strlen(buffer) - 1;
      while (whitespace(*c) && c >= buffer) *c-- = 0;

      /* Convert whitespace to a single space */
      c = buffer;
      while (whitespace(*c)) c++;
      sprintf(line+linelen, " %s", c);
      linelen += strlen(c) + 1;
    }
    else if (line[0]) {

      /* Find the value of the current header */
      c = strchr(line, ':');
      if (c) {
        c++;
        while (*c && whitespace(*c)) c++;
      }
      else {
        dprint(1, (debugfile, "read_mime_header: bogus BODY header: %s\n", line));
        break;
      }

      switch(tolower(line[0])) {
      case 'c':
        if (strncasecmp("ontent-type", line+1, sizeof("ontent-type")-1) == 0)
          parse_content_type(c, p);
        else if (strncasecmp("ontent-transfer-encoding", line+1, sizeof("ontent-transfer-encoding")-1) == 0)
          p->encoding = check_encoding(c);
        else if (strncasecmp("ontent-disposition", line+1, sizeof("ontent-disposition")-1) == 0) {
          if (strncasecmp("attachment", c, sizeof("attachment")-1) == 0)
            p->disposition = DISPATTACH;
          else
            p->disposition = DISPINLINE;

          /* Check to see if a default filename was given */
          if ((c = strchr(c, ';')) != NULL) {
	    c = skip_whitespace(c+1);
            if ((c = mutt_get_parameter("filename", (parms = parse_parameters(c)))) != 0) {
	      /*
	       * free() here because the content-type parsing routine might
	       * have allocated space if a "name=filename" parameter was
	       * specified.
	       */
	      safe_free((void **)&p->filename);
              p->filename = safe_strdup(c); 
	    }
            mutt_free_parameter(&parms);
          }
        }
        else if (strncasecmp("ontent-description", line+1, sizeof("ontent-description")-1) == 0) {
          if (!p->description) p->description = safe_strdup(c);
        }
        break;
      }

      /* Check for end of headers */
      if (buffer[0] == '\n' || (buffer[0] == '\r' && buffer[1] == '\n'))
        break;

      linelen = strlen(buffer);
      c = buffer+linelen-1;
      while (whitespace(*c) && c >= buffer) {
        *c-- = 0;
        linelen--;
      }
      strcpy(line, buffer);
    }
    else if (buffer[0] == '\n' || (buffer[0] == '\r' && buffer[1] == '\n'))
      break; /* done */
    else {
      linelen = strlen(buffer);
      c = buffer+linelen-1;
      while (whitespace(*c) && c >= buffer) {
        *c-- = 0;
        linelen--;
      }
      strcpy(line, buffer);
    }
  }
  p->offset = ftell(fp); /* Mark the start of the real data */
  if (p->type == TYPETEXT && !p->subtype) p->subtype = safe_strdup("plain");
  return(p);
}

/* Routine to parse a MESSAGE/RFC822 body */
BODY *parse_messageRFC822 (FILE *fp, BODY *parent)
{
  BODY *msg;

  msg = mutt_read_mime_header(fp);
  if (msg->type == TYPEMULTIPART)
    msg->parts = parse_multipart(fp, mutt_get_parameter("boundary", msg->parameter));
  else if (msg->type == TYPEMESSAGE)
    msg->parts = parse_messageRFC822(fp, msg);
  else
    msg->length = parent->length - (msg->offset - parent->offset);
  return(msg);
}

BODY *parse_multipart (FILE *fp, char *boundary)
{
  int blen, len, crlf = 0;
  char buffer[LONG_STRING];
  BODY *head = 0, *last = 0, *new = 0;
  short i;

  if (!boundary) {
    mutt_error("Error parsing multipart message (no boundary was found).");
    return(NULL);
  }
  blen = strlen(boundary);
  while (fgets(buffer, LONG_STRING, fp) != NULL) {
    len = strlen(buffer);

    /* take note of the line ending.  I'm assuming that either all endings
     * will use <CR><LF> or none will.
     */
    if (buffer[len-2] == '\r') crlf = 1;

    if (buffer[0] == '-' && buffer[1] == '-' && strncmp(buffer+2, boundary, blen) == 0) {
      if (last) {
        last->length = ftell(fp) - last->offset - len - 1 - crlf;
        if (last->parts && last->parts->length == 0)
          last->parts->length = ftell(fp) - last->parts->offset - len - 1 - crlf;
      }
      /* Remove any trailing whitespace, up to the length of the boundary */
      for(i=len-1; whitespace(buffer[i]) && i>blen+2; i--)
        buffer[i] = 0;
      /* Check for the end boundary */
      if (buffer[2+blen] == '-' && buffer[3+blen] == '-' && buffer[4+blen] == 0)
        break;

      new = mutt_read_mime_header(fp);
      if (head) {
        last->next = new;
        last = new;
      }
      else
        last = head = new;
    }
  }
  
  /* parse recursive MIME parts */
  last = head;
  while (last) {
    fseek(fp, last->offset, 0);
    switch (last->type) {
    case TYPEMULTIPART:
      last->parts = parse_multipart(fp, mutt_get_parameter("boundary", last->parameter));
      break;
    case TYPEMESSAGE:
      last->parts = parse_messageRFC822(fp, last);
      break;
    }
    last = last->next;
  }

  return(head);
}

static int parse_date (char *s, HEADER *h)
{
  int count = 0;
  char *t;
  int year, hour, min, sec;
  int i;

  h->zoccident = 0;

  /* kill the day of the week, if it exists. */
  if ((t = strchr(s, ',')) != 0)
    t = skip_whitespace(t+1);
  else
    t = s;

  while ((t = strtok(t, " \t")) != NULL) {
    switch (count) {
    case 0: /* day of the month */
      h->day = atoi(t);
      if ((h->day < 0) || h->day > 31) return(-1);
      break;
    case 1: /* month of the year */
      if ((i = mutt_check_month(t)) < 0) return(-1);
      h->month = i;
      break;
    case 2: /* year */
      year = atoi(t);
      if (year < 1900) year += 1900;
      h->year = year - 1971;
      break;
    case 3: /* time of day */
      if (sscanf(t, "%d:%d:%d", &hour, &min, &sec) == 3);
      else if (sscanf(t, "%d:%d", &hour, &min) == 2) {
        sec = 0;
      }
      else {
        dprint(1, (debugfile, "parse_date: could not process time format: %s\n", t));
	return(-1);
      }
      h->hours = hour;
      h->minutes = min;
      h->seconds = sec;
      break;
    case 4: /* timezone */
      if (*t == '+' || *t == '-') {
        if (*t == '-') h->zoccident = 1;
        t++;
        h->zhours = (t[0] - '0') * 10 + (t[1] - '0');
        h->zminutes = (t[2] - '0') * 10 + (t[3] - '0');
      }
      else {
	/* ad hoc support for the European MET timezone */
	if (strcasecmp(t, "MET") == 0) {
	  h->zhours = 1;
	  h->zoccident = 0;
	  if ((t = strtok((char *)0, " \t")) == (char *)0) return(-1);
	  if (strcasecmp(t, "DST") == 0) h->zhours++;
	}
        /* Most all mail should use the +HHMM format instead of character
         * strings, but here are a few defaults for the US...  */
	else if (*t == 'E') {
          h->zhours = 5;
          h->zoccident = 1;
          if (t[1] == 'D') h->zhours--;
        }
        else if (*t == 'C') {
          h->zhours = 6;
          h->zoccident = 1;
          if (t[1] == 'D') h->zhours--;
        }
        else if (*t == 'M') {
          h->zhours = 7;
          h->zoccident = 1;
          if (t[1] == 'D') h->zhours--;
        }
        else if (*t == 'P') {
          h->zhours = 8;
          h->zoccident = 1;
          if (t[1] == 'D') h->zhours--;
        }
	else if (strcmp(t, "GMT") == 0 || strcmp(t, "UTC") == 0) {
	  /* null op */
	  h->zhours = 0;
	  h->zoccident = 0;
	}
	else {
	  dprint(1,(debugfile, "parse_date(): %s: unknown timezone\n", t));
	}
      }
    }
    count++;
    t = 0;
  }
  if (count < 5) {
    return(-1);
    dprint(1,(debugfile, "parse_date(): error parsing date format, using received time\n"));
  }
  return(0);
}

/* extract the first substring that looks like a message-id */
static void extract_message_id (char *s, char *id, size_t idlen)
{
  *id = 0;
  idlen--;
  while ((s = strtok(s, " \t")) != 0) {
    if (*s == '<') {
      while (*s && *s != '>' && idlen > 0) {
	*id++ = *s++;
	idlen--;
      }
      *id++ = '>';
      *id = 0;
      return;
    }
    s = 0;
  }
}

#if 0
/*
 * This is one place where checks for buffer overflow needs to take place.
 * Currently, it tries to gather up all of the continuation lines for a
 * header before processing it.  However, this creates problems for
 * messages where there are a huge number of listed recipients...
 */
static void read_rfc822_header (FILE *fp, HEADER *hdr)
{
  char buffer[HUGE_STRING];
  char line[HUGE_STRING] = { 0 };
  char in_reply_to[SHORT_STRING] = { 0 };
  char *val;
  long content_length = -1;
  int linelen=0, line_too_long=0;
  int date_found = 0;

  if (!hdr->content) {
    hdr->content = mutt_new_body();
    hdr->content->type = TYPETEXT;
    hdr->content->subtype = safe_strdup("plain");
    hdr->content->encoding = ENC7BIT;
  }

  while (fgets(buffer, sizeof(buffer), fp) != NULL) {
    /* Remove the trailing whitespace */
    val = buffer + strlen(buffer) - 1;
    while (whitespace(*val) && val >= buffer) {
      *val = 0;
      val--;
    }

    if (buffer[0] && whitespace(buffer[0])) {
      int len;

      if (line_too_long) continue;

      /* Convert leading whitespace to a single space */
      val = skip_whitespace(buffer);

      len = strlen(val) + 1;
      if (linelen + len < sizeof(buffer)) {
      /* To delimit encoded quotes '\t' is added - pms@romance.kaist.ac.kr */
    sprintf(line+linelen, "\t %s", val); linelen++;
	linelen += len;
      }
      else
	line_too_long = 1;
    }
    else if (line[0]) {
      if ((val = strchr(line, ':')) != 0)
	val = skip_whitespace(val+1);
      switch(tolower(line[0])) {
      case 'a':
	if (strncasecmp("pparently-", line+1, 10) == 0) {
          if (strncasecmp("to:", line+11, 3) == 0)
            rfc822_parse_adrlist(&hdr->to, val, Fqdn);
          else if (strncasecmp("from:", line+11, 5) == 0)
            rfc822_parse_adrlist(&hdr->from, val, Fqdn);
	}
	break;
      case 'c':
        if (strncasecmp("ontent-", line+1, sizeof("ontent-")-1) == 0) {
          if (strncasecmp("length:", line+sizeof("ontent-"), sizeof("length:")-1)==0) {
            content_length = atoi(val);
            dprint(1,(debugfile,"read_rfc822_header: honoring Content-Length = %ld in message %d.\n", content_length, hdr->msgno));
          }
          else if (strncasecmp("transfer-encoding:", line+sizeof("ontent-"), sizeof("transfer-encoding:")-1) == 0)
            hdr->content->encoding = check_encoding(val);
          else if (strncasecmp("type:", line+sizeof("ontent-"), sizeof("type:")-1) == 0)
            parse_content_type(val, hdr->content);
          else if (strncasecmp("md5:", line+sizeof("ontent-"), sizeof("md5:")-1) == 0)
            hdr->content->md5 = safe_strdup(val);
        }
        else if (strncasecmp("c:", line+1, sizeof("c:")-1) == 0) {
          rfc822_parse_adrlist(&hdr->cc, val, Fqdn);
          rfc1522_decode_adrlist(hdr->to);
        }
        break;
      case 'd':
        if (strncasecmp("date:", line, sizeof("date:")-1) == 0) {
          if (parse_date(val, hdr) == 0)
	    date_found = 1;
#ifdef DEBUG
	  else {
	    dprint(1,(debugfile,"read_rfc822_header(): ignoring invalid date.\n"));
	  }
#endif /* DEBUG */
	}
        break;
      case 'f':
        if (strncasecmp("from:", line, sizeof("from:")-1) == 0) {
          rfc822_parse_adrlist(&hdr->from, val, Fqdn);
          rfc1522_decode_adrlist(hdr->from);
        }
        break;
      case 'i':
	if (strncasecmp("n-reply-to:", line+1, 11) == 0) {
	  /* try to find the message-id in case there is no references field */
	  extract_message_id(val, in_reply_to, sizeof(in_reply_to));
	}
	break;
      case 'l':
	if (strncasecmp("ines:", line+1, 5) == 0)
	  hdr->lines = atoi(val);
	break;
      case 'm':
        if (strncasecmp("ime-Version:", line+1, sizeof("ime-Version:")-1) == 0)
          hdr->mime = 1;
        else if (strncasecmp("essage-ID:", line+1, sizeof("essage-ID:")-1) == 0)
          hdr->message_id = safe_strdup(val);
        break;
      case 'r':
        if (strncasecmp("eply-to:", line+1, sizeof("eply-to:")-1) == 0) {
          rfc822_parse_adrlist(&hdr->reply_to, val, Fqdn);
          rfc1522_decode_adrlist(hdr->reply_to);
        }
        else if (strncasecmp("eferences:", line+1, 10) == 0)
          mutt_parse_references(&hdr->references, val);
        break;
      case 's':
        if (strncasecmp("tatus:", line+1, sizeof("tatus:")-1) == 0) {
          while (*val) {
            switch(*val){
            case 'r':
              hdr->replied = 1;
              break;
            case 'O':
	      if (option(OPTMARKOLD)) hdr->old = 1;
              break;
            case 'R':
              hdr->read = 1;
              break;
            }
            val++;
          }
        }
        else if (strncasecmp("ubject:", line+1, sizeof("ubject:")-1) == 0) {
          if (!hdr->subject && (hdr->subject = safe_strdup(val)) != 0) {
            rfc1522_decode(hdr->subject, strlen(hdr->subject), 0);
            mutt_tabs_to_spaces(hdr->subject);
          }
        }
        break;
      case 't':
        if (strncasecmp("o:", line+1, sizeof("o:")-1) == 0) {
          rfc822_parse_adrlist(&hdr->to, val, Fqdn);
          rfc1522_decode_adrlist(hdr->to);
        }
        break;
      case 'x':
        /* Be compatible with C-CLIENT so we don't lose any flags */
        if (strncasecmp("-status:", line+1, sizeof("-status:")-1) == 0) {
          while (*val) {
            switch(*val){
            case 'A':
              hdr->replied = 1;
              break;
            case 'D': /* does this ever happen? */
              hdr->deleted = 1;
              break;
	    case 'F':
	      hdr->flagged = 1;
            }
            val++;
          }
        }
        break;
      }

      /* Check for end of header. */
      if (buffer[0] == 0 || (!whitespace(buffer[0]) && strchr(buffer, ':') == 0))
        break;

      strcpy(line, buffer);
      linelen = strlen(buffer);
      line_too_long = 0;
    }
    else {
      strcpy(line, buffer);
      linelen = strlen(buffer);
      line_too_long = 0;
    }
  }
  hdr->content->hdr_offset = hdr->offset;
  hdr->content->offset = ftell(fp);

  /*
   * if we know how long this message is, either just skip over the body,
   * or if we don't know how many lines there are, count them now (this will
   * save time by not having to search for the next message marker).
   */
  if (content_length > 0) {
    if (hdr->lines > 0)
      fseek(fp, content_length, 1);
    else {
      /* count the number of lines in this message */
      while (content_length > 0) {
	if (fgets(buffer, sizeof(buffer), fp) == NULL) break;
	content_length -= strlen(buffer);
	hdr->lines++;
      }
    }
  }

  /*
   * Check for missing or invalid date.  If this is the case, convert the
   * date the message was received to UTC and use that value.
   */
  if (!date_found) {
    struct tm *when = gmtime(&hdr->received);

    hdr->year = when->tm_year - 71;
    hdr->month = when->tm_mon;
    hdr->day = when->tm_mday;
    hdr->hours = when->tm_hour;
    hdr->minutes = when->tm_min;
    hdr->seconds = when->tm_sec;
  }

  /* if no references were given, try using the In-Reply-To id (if found) */
  if (!hdr->references && in_reply_to[0])
    mutt_parse_references(&hdr->references, in_reply_to);
}
#endif

/* Note that this function is also called when new mail is appended to the
 * currently open folder, and NOT just when the mailbox is initially read.
 *
 * MMDF folders look just like MBOX folders, except that they have some
 * extra garbage around them.  Luckily we can just ignore that...
 *
 * NOTE: it is assumed that the mailbox being read has been locked before
 * this routine gets called.  Strange things could happen if it's not!
 */
int mutt_parse_mailbox (CONTEXT *ctx)
{
  struct stat sb;
  char buf[HUGE_STRING];
  char return_path[STRING];
  HEADER *curhdr;
  time_t t;
  int i, count = 0, lines = 0;
  /* if the display is not limited, make new messages visible */
  int make_visible = (ctx->msgcount == ctx->vcount);

  if (!ReadMsginc) mutt_error("Reading messages...");

  /* Save information about the folder at the time we opened it. */
  stat(ctx->folder->name, &sb);
  ctx->folder->size = sb.st_size;
  ctx->folder->mtime = sb.st_mtime;

  dprint(1,(debugfile, "mutt_parse_mailbox(): size=%d, mtime=%d\n",
	    (int)sb.st_size, (int)sb.st_mtime));

  while (fgets(buf, LONG_STRING, ctx->folder->fp) != NULL) {
    if ((t = is_from(buf, return_path, sizeof(return_path)))) {
      long loc = ftell(ctx->folder->fp) - strlen(buf);

      if (ReadMsginc && ((count % ReadMsginc) == 0))
	mutt_error("Reading message %d...", count);

      /* Save the Content-Length of the previous message */
      if (count > 0) {
#define PREVIOUS ctx->hdrs[ctx->msgcount-1]

        PREVIOUS->length = loc - PREVIOUS->offset - 1;
        PREVIOUS->content->length = loc - PREVIOUS->content->offset - 1;
	if (!PREVIOUS->lines) PREVIOUS->lines = lines ? lines - 1 : 0;
#undef PREVIOUS
      }

      if (ctx->msgcount == ctx->hdrmax) {
        /* Need to allocate more memory for structures */
        if (ctx->hdrs) {
          safe_realloc((void **)&ctx->hdrs, sizeof(HEADER *) * ((ctx->hdrmax += 25)));
	  safe_realloc((void **)&ctx->v2r, sizeof(short *) * ctx->hdrmax);
	}
        else {
          ctx->hdrs = (HEADER **)safe_malloc(sizeof(HEADER *) * ((ctx->hdrmax = 25)));
	  ctx->v2r = (short **)safe_malloc(sizeof(short *) * ctx->hdrmax);
        }

	for (i = ctx->msgcount; i < ctx->hdrmax; i++) {
	  ctx->hdrs[i] = 0;
	  ctx->v2r[i] = 0;
	}
      }
      
      curhdr = ctx->hdrs[ctx->msgcount] = mutt_new_header();
      ctx->v2r[ctx->msgcount] = (short *)safe_malloc(sizeof(short));
      *ctx->v2r[ctx->msgcount] = -1;

      curhdr->received = t;
      dprint(3, (debugfile, "parse_mailbox(): message %d received at %ld.\n",
                 ctx->msgcount, (long)curhdr->received));

      curhdr->offset = loc;
      curhdr->msgno = ctx->msgcount+1;
      curhdr->visible = make_visible;
      curhdr->env = read_rfc822_header(ctx->folder->fp, curhdr);
      if (!curhdr->read && !curhdr->old) ctx->new++;
      /*
       * check to see if there was a FROM address.  If not, use the
       * return-path we obtained from is_from()
       */
      if (!curhdr->env->from)
	rfc822_parse_adrlist(&curhdr->env->from, return_path, "@");
#ifdef _PGPPATH
      /* NOTE: this _must_ be done before the check for metamail! */
      curhdr->pgp = pgp_query(curhdr->content);
#endif /* _PGPPATH */
#ifdef METAMAIL
#ifdef _PGPPATH
      if (!curhdr->pgp)
#endif /* _PGPPATH */
        if (mutt_needs_metamail(curhdr->content)) curhdr->metamail = 1;
#endif /* METAMAIL */
      curhdr->index = ctx->msgcount++;
      count++;

      lines = 0;
    }
    else
      lines++;
  }
  
  /* Only set the content-length of the previous message if we have read more
   * than one message during _this_ invocation.  If this routine is called
   * when new mail is received, we need to make sure not to clobber what
   * previously was the last message since the headers may be sorted.
   */
  if (count > 0) {
    ctx->hdrs[ctx->msgcount-1]->length = ftell(ctx->folder->fp) - ctx->hdrs[ctx->msgcount-1]->offset - 1;
    ctx->hdrs[ctx->msgcount-1]->content->length = ftell(ctx->folder->fp) - ctx->hdrs[ctx->msgcount-1]->content->offset - 1;

    if (!ctx->hdrs[ctx->msgcount-1]->lines)
      ctx->hdrs[ctx->msgcount-1]->lines = lines ? lines - 1 : 0;
  }

  mutt_error("");
  return(0);
}

void mutt_parse_mime_message (HEADER *cur)
{
  if (cur->content->type != TYPEMESSAGE && cur->content->type != TYPEMULTIPART)
    return; /* nothing to do */
  fseek(Context->folder->fp, cur->content->offset, 0);
  if (cur->content->type == TYPEMULTIPART) {
    if (!cur->content->parts) cur->content->parts = parse_multipart(Context->folder->fp, mutt_get_parameter("boundary", cur->content->parameter));
  }
  else {
    if (!cur->content->parts)
      cur->content->parts = parse_messageRFC822(Context->folder->fp, cur->content);
  }
}

/*
 * read_rfc822_header() -- parses a RFC822 header
 *
 * Args:
 *
 * f		stream to read from
 *
 * hdr		header structure of current message (optional).  If hdr is
 *		NULL, then we are reading a postponed message, or called
 *		from ci_edit_headers() so we should keep a list of the
 *		user-defined headers.
 *
 */
ENVELOPE *read_rfc822_header (FILE *f, HEADER *hdr)
{
  char line[10240], buffer[LONG_STRING], *p;
  ENVELOPE *e = mutt_new_envelope();
  int matched;
  LIST *last = NULL;
  char in_reply_to[SHORT_STRING] = "";
  long content_length = -1;
  int date_found = 0;
  int linelen = 0;
  int len;

  if (hdr) {
    if (!hdr->content) hdr->content = mutt_new_body();
    /* reset the defaults from RFC1521 */
    hdr->content->type = TYPETEXT;
    hdr->content->subtype = safe_strdup("plain");
    hdr->content->encoding = ENC7BIT;
  }

  line[0] = 0;
  buffer[sizeof(buffer)-1] = 0; /* make sure the string is always terminated */
  while (fgets(buffer, sizeof(buffer)-1, f) != NULL) {

    /* Remove trailing whitespace from the line. */
    p = buffer + strlen(buffer) - 1;
    for (; p >= buffer; p--) {
      if (*p == '\n' || *p == ' ' || *p == '\t')
        *p = 0;
      else
        break;
    }

    if (buffer[0] == ' ' || buffer[0] == '\t') {
      /*
       * unfold the continuation line and make sure that we don't overflow the
       * buffer used to store the whole line
       */
      if (linelen < sizeof(line)) {
    /* To delimit encoded quotes '\t' is added - pms@romance.kaist.ac.kr */
	line[linelen++] = '\t';
	line[linelen] = ' ';
	linelen++;
	p = skip_whitespace(buffer);
	/* make sure we don't overwrite the buffer */
	len = strlen(p);
	if (linelen + len > sizeof(line) - 1) {
	  if ((len = sizeof(line) - linelen - 1) < 0) len = 0;
	}
	memcpy(line + linelen, p, len + 1); /* copy the terminal \0 */
	linelen += len;
      }
    }
    else if (line[0]) {
      matched = 0;

      /* find the value of this header */
      if ((p = strchr(line, ':'))) p = skip_whitespace(p+1);

      switch(tolower(line[0])) {
      case 'a':
	if (strncasecmp(line+1, "pparently-to:", 13) == 0) {
	  rfc822_parse_adrlist(&e->to, p, "@");
	  matched = 1;
	}
	else if (strncasecmp(line+1, "pparently-from:", 15) == 0) {
	  rfc822_parse_adrlist(&e->from, p, "@");
	  matched = 1;
	}
	break;
      case 'b':
        if (strncasecmp(line+1, "cc:", 3) == 0) {
          rfc822_parse_adrlist(&e->bcc, p, "@");
          matched = 1;
        }
        break;
      case 'c':
        if (strncasecmp (line+1, "c:", 2) == 0) {
          rfc822_parse_adrlist(&e->cc, p, "@");
          matched = 1;
        }
        else if (strncasecmp(line + 1, "ontent-", 7) == 0) {
          if (strncasecmp(line+8, "type:", 5) == 0) {
	    if (hdr) parse_content_type(p, hdr->content);
            matched = 1;
          }
          else if (strncasecmp(line+8, "transfer-encoding:", 18) == 0) {
	    if (hdr) hdr->content->encoding = check_encoding(p);
            matched = 1;
          }
          else if (strncasecmp(line+8, "length:", 7) == 0) {
	    content_length = atoi(p);
            matched = 1;
          }
        }
        break;
      case 'd':
	if (strncasecmp("ate:", line+1, 4) == 0) {
	  if (hdr && parse_date(p, hdr) == 0)
	    date_found = 1;
	}
	break;
      case 'f':
	if (strncasecmp(line+1, "rom:", 4) == 0) {
	  rfc822_parse_adrlist(&e->from, p, "@");
	  matched = 1;
	}
	else if (strncmp(line+1, "rom ", 4) == 0)
	  matched = 1; /* ignore the message separator */
        break;
      case 'i':
        if (strncasecmp(line+1, "n-reply-to:", 11) == 0) {
	  if (hdr) extract_message_id(p, in_reply_to, sizeof(in_reply_to));
        }
        break;
      case 'l':
	if (strncasecmp(line+1, "ines:", 5) == 0) {
	  if (hdr) hdr->lines = atoi(p);
	  matched = 1;
	}
	break;
      case 'm':
        if (strncasecmp(line+1, "ime-version:", 12) == 0) {
	  if (hdr) hdr->mime = 1;
          matched = 1;
        }
	else if (strncasecmp(line+1, "essage-id:", 10) == 0) {
	  if (hdr) hdr->message_id = safe_strdup(p);
	}
        break;
      case 'r':
        if (strncasecmp(line+1, "eferences:", 10) == 0) {
	  mutt_parse_references(&e->references, p);
          matched = 1;
        }
	else if (strncasecmp(line+1, "eply-to:", 8) == 0) {
	  if (hdr) rfc822_parse_adrlist(&e->reply_to, p, "@");
	}
        break;
      case 's':
        if (strncasecmp (line+1, "ubject:", 7) == 0) {
          if (!e->subject) e->subject = safe_strdup(p);
          matched = 1;
        }
        else if (strncasecmp(line+1, "ender:", 6) == 0) {
          rfc822_parse_adrlist(&e->sender, p, "@");
          matched = 1;
        }
	else if (strncasecmp(line+1, "tatus:", 6) == 0) {
	  if (hdr) {
	    while (*p) {
	      switch(*p) {
	      case 'r':
		hdr->replied = 1;
		break;
	      case 'O':
		if (option(OPTMARKOLD)) hdr->old = 1;
		break;
	      case 'R':
		hdr->read = 1;
		break;
	      }
	      p++;
	    }
	  }
	  matched = 1;
	}
        break;
      case 't':
        if (strncasecmp (line+1, "o:", 2) == 0) {
          rfc822_parse_adrlist(&e->to, p, "@");
          matched = 1;
        }
        break;
      case 'x':
	if (strncasecmp(line+1, "-status:", 8) == 0) {
	  if (hdr) {
	    while (*p) {
	      switch (*p) {
	      case 'A':
		hdr->replied = 1;
		break;
	      case 'D':
		hdr->deleted = 1;
		break;
	      case 'F':
		hdr->flagged = 1;
		break;
	      default:
		break;
	      }
	      p++;
	    }
	  }
	  matched = 1;
	}
	break;
      default:
	break;
      }
      /*
       * if hdr==NULL, then we are using this to parse either a postponed
       * message, or a outgoing message (edit_hdrs), so we want to keep
       * track of the user-defined headers
       */
      if (!matched && !hdr) {
	if (last) {
	  last->next = mutt_new_list();
	  last = last->next;
	}
	else
	  last = e->userhdrs = mutt_new_list();
	last->data = safe_strdup(line);
      }
      if (buffer[0] == 0) break; /* end of headers */
      linelen = strlen(buffer);
      memcpy(line, buffer, linelen + 1);
    }
    else {
      linelen = strlen(buffer);
      memcpy(line, buffer, linelen + 1);
    }
  }

  if (hdr) {
    hdr->content->hdr_offset = hdr->offset;
    hdr->content->offset = ftell(f);

    /*
     * if we know how long this message is, either just skip over the body,
     * or if we don't know how many lines there are, count them now (this will
     * save time by not having to search for the next message marker).
     */
    if (content_length > 0) {
      if (hdr->lines > 0)
	fseek(f, content_length, 1);
      else {
	dprint(1,(debugfile,"read_rfc822_header(): no Lines field found, counting...\n"));
	/* count the number of lines in this message */
	while (content_length > 0) {
	  if (fgets(buffer, sizeof(buffer), f) == NULL) break;
	  content_length -= strlen(buffer);
	  hdr->lines++;
	}
      }
    }

    /*
     * Check for missing or invalid date.  If this is the case, convert the
     * date the message was received to UTC and use that value.
     */
    if (!date_found) {
      struct tm *when = gmtime(&hdr->received);

      hdr->year = when->tm_year - 71;
      hdr->month = when->tm_mon;
      hdr->day = when->tm_mday;
      hdr->hours = when->tm_hour;
      hdr->minutes = when->tm_min;
      hdr->seconds = when->tm_sec;

      dprint(1,(debugfile,"read_rfc822_header(): no date found, using received time from msg separator\n"));
    }

    /* if no references were given, try using the In-Reply-To id (if found) */
    if (!e->references && in_reply_to[0]) {
      mutt_parse_references(&e->references, in_reply_to);
      dprint(3,(debugfile,"read_rfc822_header(): no References found, using In-Reply-To\n"));
    }
    
    /* do RFC1522 decoding */
    rfc1522_decode_adrlist(e->from);
    rfc1522_decode_adrlist(e->to);
    rfc1522_decode_adrlist(e->cc);
    if (e->subject) rfc1522_decode(e->subject, strlen(e->subject), 0);
  }

  return e;
}
