/* Copyright (C) 1996 Park Myeong Seok <pms@romance.kaist.ac.kr>.
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 

#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include "muttlib.h"
#include "rfc_gen.h"

int CompatKorean(const char *charset)
{
   if( !strcasecmp(charset, "euc-kr") || !strcasecmp(charset, "iso-2022-kr")
       || !strcasecmp(charset, "korean") ) return 1;
   return 0;
}

/*
**  ISO2022KR decoder 
**     - flexibility for RFC1557
*/
int DecodeISO2022KR(STATE *s, long len)
{
   register int c, shift=0, b_iso=0, iso_2022_kr=0;    

   dprint(3,(debugfile,"DecodeISO2022KR(): entering.\n"));
   if(s->prefix) state_puts(s->prefix, s);
   while(len > 0) {
      c = fgetc(s->fpin); len--;
      if(c==0x1B) { /* Ignored ISO-2022-KR flag! */
         if((c=fgetc(s->fpin))==0x24) if((c=fgetc(s->fpin))==0x29)
         if((c=fgetc(s->fpin))==0x43) { len -= 3; b_iso = 1; continue; }
      }
      if(c==0x0A && b_iso) continue;
      else if(c==0x0E) shift = 1;
      else if(c==0x0F) shift = 0;
      else if(shift) {
         if(!iso_2022_kr) iso_2022_kr = 1;
         if(c==0x20 || c==0x09 || c==0x0A || c==0x0D) {
            state_putc(c, s);
            if(c==0x0A && s->prefix) state_puts(s->prefix, s);
         }
         else state_putc(c+0x80, s);
      } 
      else {
         state_putc(c, s);
         if(c==0x0A && s->prefix) state_puts(s->prefix, s);
      }
      b_iso = 0;
   }
   dprint(3,(debugfile,"DecodeISO2022KR(): exit.\n"));
   return iso_2022_kr;
}

/*
** ISO-2022-KR encoding 
*/
void EncodeISO2022KR(FILE *fin, FILE *fout)
{
   register int c, kor=0, ksc=0;
   unsigned char buf[LONG_STRING], *s;
    
   while((s = fgets(buf, sizeof(buf), fin)) != NULL) {
      if(!ksc) {
         while(*s++) if(*s > 0x80) { ksc = 1; s = buf; break; }
         if(ksc) fputs("\033$)C", fout);
         else fputs(buf, fout);
      } 
      if(ksc) while((c=*s++) != EOF && c) {
         if(c>0x80) {
            if(!kor) fputc(0x0E, fout);
            kor = 1; fputc(c-0x80, fout);
         }
         else {
            if(kor) fputc(0x0F, fout);
            kor = 0; fputc(c, fout);
         }
      }
   }
   if(ksc & kor) fputc(0x0F, fout);
}

static char *b64_alphabet = 
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";

/*
** Copyright (C) 1992 Ienup Sung.
**    - simplified
*/
void DecodeBase64(char *ob, const char *ib)
{
   register unsigned long bm;
   register short int cnt;

   for(bm=cnt=0; *ib!=0x00; ib++) {
      bm = (*ib==0x2B)?  (bm << 6) | 0x3E:
           (*ib==0x2F)?  (bm << 6) | 0x3F:
           isdigit(*ib)? (bm << 6) | (0x34 + *ib - 0x30):
           isupper(*ib)? (bm << 6) | (*ib - 0x41):
           islower(*ib)? (bm << 6) | (0x1A + *ib - 0x61):
           (bm << 6) | 0;
      if(++cnt >= 4) {
           *ob++ = (char)((bm & 0x0FF0000) >> 16);
           *ob++ = (char)((bm & 0x000FF00) >> 8);
           *ob++ = (char) (bm & 0x00000FF);
           bm = cnt = 0;
      }
   }
   *ob = 0;
}

/*
** Copyright (C) 1992 Ienup Sung.
**    - simplified
*/
void EncodeBase64(char *ob, const char *ib)
{
   register unsigned long bm;
   register short int i, cnt;

   while (*ib) {
      for(i=cnt=0, bm=0; i<3; i++) {
         if(*ib) {
            bm = (bm<<8) | (unsigned long)(*ib++ & 0x0FF);
            cnt++;
         }
         else bm = (bm<<8) | (unsigned long)0;
      }
      *ob++ = b64_alphabet[(0x0FC0000 & bm) >> 18];
      *ob++ = b64_alphabet[(0x003F000 & bm) >> 12];
      *ob++ = b64_alphabet[(cnt>=2)? (0x0000FC0 & bm) >> 6: 64];
      *ob++ = b64_alphabet[(cnt==3)? (0x000003F & bm): 64];
   }
   *ob = 0;
}

static int HeaderCode(char c)
{
   return ((c==0x42) | (c==0x62))? 4: ((c==0x51) | (c==0x71))? 3: 0;
}

/*
**  Quoted Printable decoding
*/
static int DeQP(const char *s)
{
   register int i, hc, hb;
    
   for(i=hc=0; i<2; i++) {
      hb = (s[i] > 0x2F && s[i] < 0x3A)? s[i] - 0x30:
           (s[i] > 0x40 && s[i] < 0x47)? s[i] - 0x37:
           /* ignore lower case character!
           (s[i] > 0x60 && s[i] < 0x67)? s[i] - 0x57:
           */
           -1;
      if(hb<0) return 0;
      hc = (i)? hc | hb: hb << 4; 
   }
   return hc;
}

/*
**  Quoted Printable decoding
*/
void DecodeQP(char *d, const char *s)
{
   char c, h[2];
   register int i, j, sl = strlen(s);

   for(i=0; i<sl; i++) {
      if(s[i]==0x3D && i<sl-2) {
         for(j=0; j<2; j++) h[j] = s[i+1+j];
         if(!(c=DeQP(h))) *d++ = 0x3D;
         else { i+= 2; *d++ = c; }
      }
      else if(s[i] == 0x5F) *d++ = 0x20;
      else *d++ = s[i];
   }
   *d = 0;
}

/*
**  Header B-encoding
*/
void rfc1522_B_encode_string(char *d, unsigned char *s, int l)
{
   unsigned char *e=d;
   char charset[32];
   register int i, count, nel, ewl, ewl0, ewl1, csl, ch8;

   /* First check to see if there are any 8-bit characters */
   for(i=count=nel=0; !count && i<strlen(s); i++) {
      if(s[i] & 0x80) count++;
      else if(count==0 && s[i]==' ') nel = i;
   }
   if(!count) { strfcpy(d, s, l); return; }

   if(CompatKorean(Charset)) strcpy(charset, "EUC-KR");
   else strcpy(charset, Charset);
   csl = strlen(charset);
   if(nel) {
      for(i=0; i<nel+1; i++) *e++ = s[i];
      s += nel + 1;
   }
   ewl0 = ((67 - 13 - nel - csl)/4)*3 - 2;
   ewl1 = ((67 - csl)/4)*3 - 2;

   for(count=0; *s; ) { 
      unsigned char ib[STRING]="", ob[STRING]="", *h = ib;
      ewl = (count)? ewl1: ewl0;
      for(i=0; *s && i<ewl; i++) {
         ch8 = ((*s) & 0x80)? ch8++: 0;
         *h++ = *s++;
      }
      /* Do not cut a double-byte pair! */
      if(ch8%2) { *(--h) = 0; --s; }
      else *h = 0;
      EncodeBase64(ob, ib);
      sprintf(e, "=?%s?B?%s?=", charset, ob);
      e += strlen(ob) + 7 + csl;
      if(*s) { sprintf(e, "\n "); e += 2; count++; }
   }
   *e = 0;
}

/*
**  Email Header Decoding
**      QP=3, B64=4
*/
static int DeHeader(char *d, int code, const char *s)
{
    char ts[LONG_STRING];
    register int i, sl;

    if(code==3) DecodeQP(ts, s);
    if(code==4) DecodeBase64(ts, s);
    sl = strlen(ts);
    for(i=0; i<sl; i++) *d++ = ts[i];
    return sl;
}

/*
**  Email Header Decoding
**      QP=3, B64=4
*/
void DecodeHeader(char *d, const char *s)
{
   char ts[LONG_STRING];
   register int i, j, c, code, sl, fail;

   sl = strlen(s);
   for(i=c=code=0; i<sl; i++) {
      if(i<sl-1 && s[i]==0x3D && s[i+1]==0x3F) {
         fail = 0;
         if(i<sl-3 && s[i+3]==0x3F) {
            if(!code) code = HeaderCode(s[i+2]);
            if(code) {
               for(j=0;j<sl-i-4 && s[i+4+j]!=0x3F;j++);
               if(j>sl-i-4) fail = 1;
               else c = i + 5 + j;
            }
            else fail = 1;
         }
         else {
            for(j=0; j<sl-i-2 && s[i+2+j]!=0x3F; j++);
            if(j<sl-i-3) {
               if(!code) code = HeaderCode(s[i+3+j]); 
               if(code && j<sl-i-4 && s[i+4+j]==0x3F) c = i + 5 + j;
               else fail = 1;
            }
            else fail = 1;
         }
         if(!fail) {
            for(j=0; c<sl && s[c]!=0x3F; j++, c++) ts[j] = s[c];
            if(c<sl-1 && s[c+1]==0x3D) {
               c++;
               ts[j] = 0; d += DeHeader(d, code, ts);
            }
            else fail = 1;
         }
         if(fail) { for(j=0; j<2; j++) *d++ = s[i+j]; i++; }
         else i = c;
      } 
      else *d++ = s[i]; 
   }
   *d = 0;
}
