/*
 * Copyright (C) 1996 Michael R. Elkins <me@cs.hmc.edu>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include "muttlib.h"
#include "rfc822.h"
#include "rfc1522.h"
#include "rfc_gen.h"

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>

#define DISPOSITION(X) X==DISPATTACH?"attachment":"inline"

extern char *tspecials;

static char base64_chars[64] = {
  'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O',
  'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd',
  'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's',
  't', 'u', 'v', 'w', 'x', 'y', 'z', '0', '1', '2', '3', '4', '5', '6', '7',
  '8', '9', '+', '/'
};

static void encode_quoted (FILE * fin, FILE *fout, int istext)
{
  int c, linelen = 0;
  char line[77], savechar;

  while ((c = fgetc(fin)) != EOF) {
    /* Escape lines that begin with "the message separator". */
    if (linelen == 5 && strncmp("From ", line, 5) == 0) {
      line[linelen-1] = '=';
      line[linelen++] = '2';
      line[linelen++] = '0';
    }

    /* Wrap the line if needed. */
    if (linelen == 76 && ((istext && c != '\n') || !istext)) {
      /* If the last character is "quoted", then be sure to move all three
       * characters to the next line.  Otherwise, just move the last
       * character...
       */
      if (line[linelen-3] == '=') {
        line[linelen-3] = 0;
        fputs(line, fout);      
        fputs("=\n", fout);
        line[linelen] = 0;
        line[0] = '=';
        line[1] = line[linelen-2];
        line[2] = line[linelen-1];
        linelen = 3;
      }
      else {
        savechar = line[linelen-1];
        line[linelen-1] = '=';
        line[linelen] = 0;
        fputs(line, fout);
        fputc('\n', fout);
        line[0] = savechar;
        linelen = 1;
      }
    }

    if (c == '\n' && istext) {
      /* Check to make sure there is no trailing space on this line. */
      if (line[linelen-1] == ' ' || line[linelen-1] == '\t') {
        if (linelen < 74) {
          sprintf(line+linelen-1, "=%2.2X", line[linelen-1]);
          fputs(line, fout);
        }
        else {
          int savechar = line[linelen-1];

          line[linelen-1] = '=';
          line[linelen] = 0;
          fputs(line, fout);
          fprintf(fout, "\n=%2.2X", savechar);
        }
      }
      else {
        line[linelen] = 0;
        fputs(line, fout);
      }
      fputc('\n', fout);
      linelen = 0;
    }
    else if (c != 9 && (c < 32 || c > 126 || c == '=')) {
      /* Check to make sure there is enough room for the quoted character.
       * If not, wrap to the next line.
       */
      if (linelen > 73) {
        line[linelen++] = '=';
        line[linelen] = 0;
        fputs(line, fout);
        fputc('\n', fout);
        linelen = 0;
      }
      sprintf(line+linelen,"=%2.2X", c);
      linelen += 3;
    }
    else {
      /* Don't worry about wrapping the line here.  That will happen during
       * the next iteration when I'll also know what the next character is.
       */
      line[linelen++] = c;
    }
  }

  /* Take care of anything left in the buffer */
  if (linelen > 0) {
    if (line[linelen-1] == ' ' || line[linelen-1] == '\t') {
      /* take care of trailing whitespace */
      if (linelen < 74)
        sprintf(line+linelen-1, "=%2.2X", line[linelen-1]);
      else {
        savechar = line[linelen-1];
        line[linelen-1] = '=';
        line[linelen] = 0;
        fputs(line, fout);
        fputc('\n', fout);
        sprintf(line, "=%2.2X", savechar);
      }
    }
    else
      line[linelen] = 0;
    fputs(line, fout);
  }
}

static void encode_base64 (FILE * fin, FILE *fout, int istext)
{
  int c1, c2, c3, ch;
  int insert_newline = 0;
  int linelen = 0;

  for (;;) {
    if (istext) {
      if (insert_newline) {      
        c1 = '\n';
        insert_newline = 0;
  
        c2 = fgetc(fin);
        if (c2 == '\n') {
          c2 = '\r';
          c3 = '\n';
        }
        else {
          c3 = fgetc(fin);
          if (c3 == '\n') {
            c3 = '\r';
            insert_newline = 1;
          }
        }
      }
      else {
        c1 = fgetc(fin);
        if (c1 == '\n') {
          c1 = '\r';
          c2 = '\n';
          c3 = fgetc(fin);
          if (c3 == '\n') {
            c3 = '\r';
            insert_newline = 1;
          }
        }
        else {
          c2 = fgetc(fin);
          if (c2 == '\n') {
            c2 = '\r';
            c3 = '\n';
          }
          else {
            c3 = fgetc(fin);
            if (c3 == '\n') {
              c3 = '\r';
              insert_newline = 1;
            }
          }
        }
      }
    }
    else { /* !istext */
      if ((c1 = fgetc(fin)) == EOF)
        break;
      c2 = fgetc(fin);
      c3 = fgetc(fin);
    }

    if (linelen + 4 >= 76) {
      fputc('\n', fout);
      linelen = 0;
    }

    ch = c1 >> 2;
    fputc(base64_chars[ch], fout);

    if (c2 != EOF) {
      ch = ((c1 & 0x3) << 4) | (c2 >> 4);
      fputc(base64_chars[ch], fout);
    }
    else {
      ch = (c1 & 0x3) << 4;
      fputc(base64_chars[ch], fout);
      fputs("==", fout);
      break;
    }

    if (c3 != EOF) {
      ch = ((c2 & 0xf) << 2) | (c3 >> 6);
      fputc(base64_chars[ch], fout);
    }
    else {
      ch = (c2 & 0xf) << 2;
      fputc(base64_chars[ch], fout);
      fputc('=', fout);
      break;
    }

    ch = c3 & 0x3f;
    fputc(base64_chars[ch], fout);

    linelen += 4;
  }

  fputc('\n', fout);
}

/* Don't print a Content-Type header if everything is the default. */
static int needs_content_type (BODY * a)
{
  char *p;

  if (a->type != TYPETEXT) return 1;
  if (strcasecmp(a->subtype, "plain") != 0) return 1;
  p = mutt_get_parameter("charset", a->parameter);
  if (!p) return 0;
  if (strcasecmp(p, "us-ascii") == 0) return 0;
  return 1;
}

int mutt_write_mime_header (BODY *a, FILE *f)
{
  PARAMETER *p;
  char buffer[STRING];
  char *t;

  if (needs_content_type (a)) {
    sprintf(buffer, "Content-Type: %s/%s", TYPE(a->type), a->subtype);
    fputs(buffer, f);
    if (a->parameter) {
      int len = strlen(buffer);

      p = a->parameter;
      while (p) {
        sprintf(buffer, "%s=", p->attribute);
        rfc822_cat(buffer, p->value, tspecials);
        if (len + strlen(buffer) + 2 > 76) {
          fprintf(f, ";\n\t%s", buffer);
          len = strlen(buffer) + 8;
        }
        else {
          len += strlen(buffer);
          fprintf(f, "; %s", buffer);
        }
        p = p->next;
      }
    }
    fputc('\n', f);
  }
  if (a->encoding != ENC7BIT)
    fprintf(f, "Content-Transfer-Encoding: %s\n", ENCODING(a->encoding));
  if (a->description)
    fprintf(f, "Content-Description: %s\n", a->description);
  if (a->use_disp && (a->disposition == DISPATTACH || a->filename)) {
    fprintf(f, "Content-Disposition: %s", DISPOSITION(a->disposition));
    if (a->filename) {
      /* Strip off the leading path... */
      t = strrchr(a->filename, '/');
      if (t) t++;
      else t = a->filename;
      
      buffer[0] = 0;
      rfc822_cat(buffer, t, tspecials);
      fprintf(f, "; filename=%s", buffer);
    }
    fputc('\n', f);
  }
  /* Do NOT add the terminator here!!! */
  return 0;
}

/*
**  Support nation specific RFCs - pms@romance.kaist.ac.kr
*/
int mutt_write_mime_body (BODY *a, FILE *f)
{
  char boundary[SHORT_STRING], buffer[LONG_STRING];
  FILE *fpin;
  BODY *t;
  PARAMETER *p;
  long len;

  if (a->type == TYPEMULTIPART) {
    /* First, find the boundary to use */
    p = a->parameter;
    while (p) {
      if (strcmp("boundary", p->attribute) == 0) {
        strcpy(boundary, p->value);
        break;
      }
      p = p->next;
    }

    t = a->parts;
    while (t) {
      fprintf(f, "\n--%s\n", boundary);
      mutt_write_mime_header(t, f);
      fputc('\n', f);
      if (mutt_write_mime_body(t, f) == -1) return -1;
      t = t->next;
    }
    fprintf(f, "\n--%s--\n", boundary);
    return 0;
  }

#ifdef _PGPPATH
  /* This is pretty gross, but it's the best solution for now... */
  if (a->type == TYPEAPPLICATION && strcmp (a->subtype, "pgp-encrypted") == 0) {
    fputs ("Version: 1\n", f);
    return 0;
  }
#endif /* _PGPPATH */

  if (Context && Context->folder && Context->folder->name &&
      strncmp(a->filename, Context->folder->name, strlen(Context->folder->name)) == 0) {
    /* Forwarded message. */
    short msgno, inheader = 1;
    long bytes;
    char *c = strrchr(a->filename, ':');

    if (!c) goto recover; /* include the _whole_ file, not just a msg. */
    c++;
    fpin = Context->folder->fp;
    msgno = atoi(c);
    fseek(fpin, Context->hdrs[msgno]->offset, 0);
    bytes = Context->hdrs[msgno]->length;

    /* The first line of the message will be the mailbox separator, which
     * is _not_ an RFC822 header, so don't include it... */
    if (fgets(buffer, sizeof(buffer), fpin) == NULL) return -1;
    bytes -= strlen(buffer);
    while (bytes > 0) {
      if (fgets(buffer, sizeof(buffer), fpin) == NULL) return -1;
      len = strlen(buffer);
      bytes -= len;
      if (inheader) {
        /* Don't include of the lines added by Mutt. */
        if (strncmp("Status:", buffer, sizeof("Status:")-1) == 0 ||
            strncmp("X-Status:", buffer, sizeof("X-Status:")-1) == 0 ||
            strncmp("Content-Length:", buffer, 15) == 0)
          continue;
        if (buffer[0] == '\n') inheader = 0;
      }
      fputs(buffer, f);
    }
  }
  else {
  recover:
    if ((fpin = fopen(a->filename, "r")) == NULL) {
      dprint(1,(debugfile, "write_mime_body: %s no longer exists!\n",a->filename));
      return -1;
    }
    
    if (a->encoding == ENCQUOTEDPRINTABLE)
      encode_quoted(fpin, f, mutt_is_text_type(a->type, a->subtype));
    else if (a->encoding == ENCBASE64)
      encode_base64(fpin, f, mutt_is_text_type(a->type, a->subtype));
    else {
      if (!(a->encoding == ENC7BIT && mutt_encode_own_rfc(fpin, f))) {
        while ((len = fread(buffer, 1, sizeof(buffer), fpin)) > 0)
           fwrite(buffer, 1, len, f);
      }
    }
    fclose(fpin);
  }
  return 0;
}

#define BOUNDARYLEN 16
char *mutt_generate_boundary (void)
{
  char *rs = (char *)safe_malloc(BOUNDARYLEN+1);
  char *p = rs;
  const char bchars[] = {
    'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
    'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', 'A', 'B',
    'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
    'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '1', '2', '3', '4',
    '5', '6', '7', '8', '9', '0', '+', '='
  };
  int i;

  rs[BOUNDARYLEN] = 0;
  for (i=0;i<BOUNDARYLEN;i++)
    *p++ = bchars[LRAND() % sizeof(bchars)];
  *p = 0;
  return(rs);
}

/* analyze the contents of a file to determine which MIME encoding to use */
CONTENT *mutt_get_content_info (const char *fname)
{
  CONTENT *info;
  FILE *fp;
  int ch, from=0, whitespace=0, dot=0, linelen=0;

  if ((fp = fopen(fname, "r")) == 0) return(0);
  info = (CONTENT *)safe_malloc(sizeof(CONTENT));
  memset(info, 0, sizeof(CONTENT));
  while ((ch = fgetc(fp)) != EOF) {
    linelen++;
    if (ch == '\n') {
      if (whitespace) info->space=1;
      if (dot) info->dot=1;
      if (linelen > info->linemax) info->linemax = linelen;
      whitespace=0;
      linelen=0;
      dot=0;
    }
    else if (ch == '\r') {
      if ((ch = fgetc(fp)) == EOF) {
        info->binary = 1;
        break;
      }
      else if (ch != '\n') {
        info->binary = 1;
	ungetc(ch, fp);
	continue;
      }
      else {
        if (whitespace) info->space = 1;
	if (dot) info->dot = 1;
        if (linelen > info->linemax) info->linemax = linelen;
        whitespace=0;
	dot=0;
        linelen=0;
      }
    }
    else if (ch & 0x80)
      info->hibin++;
    else if (ch == '\t') {
      info->ascii++;
      whitespace++;
    }
    else if (ch < 32 || ch == 127)
      info->lobin++;
    else {
      if (linelen == 1) {
        if (ch == 'F')
          from=1;
        else
          from=0;
        if (ch == '.')
          dot=1;
        else
          dot=0;
      }
      else if (from) {
        if (linelen == 2 && ch != 'r') from=0;
        else if (linelen == 3 && ch != 'o') from=0;
        else if (linelen == 4 && ch != 'm') from=0;
        else if (linelen == 5) {
          if (ch == ' ') info->from=1;
          from = 0;
        }
      }
      if (ch == ' ') whitespace++;
      info->ascii++;
    }
    if (linelen > 1) dot=0;
    if (ch != ' ' && ch != '\t') whitespace=0;
  }
  return(info);
}

/*
** Support nation specific RFCs - pms@romance.kaist.ac.kr
*/
char *set_text_charset (CONTENT *info)
{
  int allow_8bit = option(OPTALLOW8BIT);

  if (strcasecmp(Charset, "us-ascii") == 0) {
    if (info->hibin != 0)
      return "unknown-8bit";
  }
  else if (info->hibin == 0) {
    if (mutt_compat_charset(Charset))
      return "us-ascii";
  }
  return (rfc_own_charset(allow_8bit, info->hibin));
}

/* 
 * given a file with extension ``s'', see if there is a registered MIME type.
 * returns the major MIME type, and copies the subtype to ``d''.
 * First look for ~/.mime.types, then look in a system mime.types if we can
 * find one.
 */
static int lookup_mime_type (char *d, const char *s)
{
  FILE *f = NULL;
  char *p, *q, buf[LONG_STRING];
  char *c;
  int count=0;

  *d = 0;

read_file:

  switch (count) {
  case 0:
    sprintf(buf, "%s/.mime.types", Homedir);
    f = fopen(buf, "r");
    break;
  case 1:
    f = fopen(MIMETYPES, "r");
    break;
  default:
    return(-1);
  }

  count++;
  if (f == NULL) goto read_file;

  while (fgets(buf, sizeof(buf)-1, f) != NULL) {
    /* weed out any comments */
    if ((p = strchr(buf, '#'))) *p = 0;

    /* remove any leading space. */
    q = buf;
    while (*q && (*q == ' ' || *q == '\t')) q++;

    /* position on the next field in this line */
    if ((p = strpbrk(q, " \t")) == NULL) continue;
    *p = 0;
    p = skip_whitespace(p+1);

    /* cycle through the file extensions */
    while ((p = strtok(p, " \t\n"))) {
      if (strcasecmp(s, p) == 0) {
	/* get the content-type */

	if ((c = strchr(q, '/')) == NULL) {
	  /* malformed line, just skip it. */
	  break;
	}
	*c++ = 0;

	while (*c && *c != ' ' && *c != '\t' && *c != '\n' && *c != '\r')
	  *d++ = *c++;
	*d = 0;

	fclose(f);
	return(mutt_check_mime_type(q));
      }
      p = NULL;
    }
  }
  fclose(f);
  f = NULL;
  goto read_file;
}

/*
**  Support nation specific RFCs - pms@romance.kaist.ac.kr
*/
BODY *mutt_make_attach (const char *path)
{
  char *p, buf[SHORT_STRING];
  int n;
  BODY *att = mutt_new_body();
  CONTENT *info;
  
  att->filename = safe_strdup(path);
  if ((p = strrchr(path, '.'))) {
    p++;

    /* check to see if there is an entry in mime.types */
    n = lookup_mime_type(buf, p);

    if (n != -1) {
      att->type = n;
      att->subtype = safe_strdup(buf);
    }
  }
  info = mutt_get_content_info(path);
  if (!att->subtype) {
    if (info->lobin == 0 || (info->lobin + info->hibin + info->ascii)/ info->lobin > 4) {
      /* Statistically speaking, there should be more than 25% "lobin" 
       * chars if this is really a binary file... */
      att->type = TYPETEXT;
      att->subtype = safe_strdup("plain");
      att->parameter = mutt_new_parameter();
      att->parameter->attribute = safe_strdup("charset");
      att->parameter->value = safe_strdup(set_text_charset(info));
    }
    else {
      att->type = TYPEAPPLICATION;
      att->subtype = safe_strdup("octet-stream");
    }
  }
  
  /* now determine which Content-Transfer-Encoding to use */
  if (att->type != TYPETEXT || info->binary || info->linemax > 990) {
    /* Determine which encoding is smaller  */
    if (1.33 * (float)(info->lobin+info->hibin+info->ascii) < 3.0 * (float) (info->lobin + info->hibin) + (float)info->ascii)
      att->encoding = ENCBASE64;
    else
      att->encoding = ENCQUOTEDPRINTABLE;
  }
  else if (info->hibin)
    att->encoding = option(OPTALLOW8BIT) ? ENC8BIT : 
                    (rfc_own_content_encoding())? ENC7BIT :
                    ENCQUOTEDPRINTABLE;
  else
    att->encoding = ENC7BIT;
  
#ifdef _PGPPATH
  /*
   * save the info in case this message is signed.  we will want to do Q-P
   * encoding if any lines begin with "From " so the signature won't be munged,
   * for example.
   */
  att->content = info;
#else
  safe_free((void **)&info);
#endif /* _PGPPATH */

  return(att);
}

static int get_toplevel_encoding (BODY *a)
{
  int e = ENC7BIT;

  while (a) {
    if (a->encoding == ENCBINARY) return(ENCBINARY);
    else if (a->encoding == ENC8BIT) {
      if (e == ENC7BIT) e = ENC8BIT;
    }
    a = a->next;
  }
  return(e);
}

BODY *mutt_make_multipart (BODY *b)
{
  BODY *new;

  new = mutt_new_body();
  new->type = TYPEMULTIPART;
  new->subtype = safe_strdup("mixed");
  new->encoding = get_toplevel_encoding(b);
  new->parameter = mutt_new_parameter();
  new->parameter->attribute = safe_strdup("boundary");
  new->parameter->value = mutt_generate_boundary();
  new->use_disp = 0;  
  new->parts = b;

  return new;
}

/*
 * wrapper around mutt_write_address() so we can handle very large
 * recipient lists without needing a huge temporary buffer in memory
 */
static void write_address_list (ADDRESS *adr, FILE *fp)
{
  ADDRESS *tmp;
  char buf[LONG_STRING];
  int count=0;
  int linelen=4; /* assume the line begins with "To: " or "Cc: " */
  int len;

  while (adr) {
    tmp = adr->next;
    adr->next = 0;
    buf[0] = 0;
    mutt_write_address(buf, sizeof(buf), adr);
    len = strlen(buf);
    if (count && linelen + len > 74) {
      if (count) {
	linelen = len + 8; /* tab is usually about 8 spaces... */
	fputs(",\n\t", fp);
      }
    }
    else {
      if (count) {
	fputs(", ", fp);
	linelen += 2;
      }
      linelen += len;
    }
    fputs(buf, fp);
    adr->next = tmp;
    adr = adr->next;
    count++;
  }
  fputc('\n', fp);
}

/* Note: all RFC1522 encoding should be done outside of this routine, except
 * for the "real name."  This will allow this routine to be used more than
 * once, if necessary.
 *
 * mode == 1  => "lite" mode (used for edit_hdrs)
 * mode == 0  => normal mode.  write full header + MIME headers
 * mode == -1 => write just the envelope info (used for postponing messages)
 */

int mutt_write_rfc822_header (FILE *fp, ENVELOPE *env, BODY *attach, int mode)
{
  char buffer[LONG_STRING];
  LIST *tmp = env->userhdrs;

  if (mode == 0) fputs(mutt_make_date(buffer), fp);

  if (env->from) {
    buffer[0] = 0;
    mutt_write_address(buffer, sizeof(buffer), env->from);
    fprintf(fp, "From: %s\n", buffer);
  }

  if (env->to) {
    fputs("To: ", fp);
    write_address_list(env->to, fp);
  }
  else if (mode > 0)
    fputs("To: \n", fp);

  if (env->cc) {
    fputs("Cc: ", fp);
    write_address_list(env->cc, fp);
  }
  else if (mode > 0)
    fputs ("Cc: \n", fp);

  if (env->bcc) {
    fputs("Bcc: ", fp);
    write_address_list(env->bcc, fp);
  }
  else if (mode > 0)
    fputs ("Bcc: \n", fp);

  if (env->subject)
    fprintf(fp, "Subject: %s\n", env->subject);
  else if (mode == 1)
    fputs("Subject: \n", fp);

  if (mode <= 0) {
    if (env->references) {
      LIST *p = env->references;

      fputs("References:", fp);
      while (p) {
	fprintf(fp, " %s", p->data);
	p = p->next;
      }
      fputc('\n', fp);
    }
  }

  if (mode == 0) {
#ifndef NO_XMAILER
    /* Add a vanity header */
    fprintf(fp, "X-Mailer: %s\n", Mutt_version);
#endif

    /* Add the MIME headers */
    fputs("Mime-Version: 1.0\n", fp);
    mutt_write_mime_header(attach, fp);
  }

  /* Add any user defined headers */
  while (tmp) {
    fputs(tmp->data, fp);
    fputc('\n', fp);
    tmp = tmp->next;
  }

  return(ferror (fp) == 0 ? 0 : -1);
}

/*
**  Allow_8bit more strictly - pms@romance.kiast.ac.kr
*/
int mutt_send_message (ENVELOPE *en,
		       BODY *attach,
		       const char *fcc, /* where to save a copy of this msg */
		       int batchmode)
{
  char tempfile[_POSIX_PATH_MAX];
  char buffer[STRING];
  char cmd[STRING];
  FILE *tempfp;

    /* Follow allow_8bit setting more strictly!! */
    if (attach->encoding != ENC8BIT) {
  /* Take care of 8-bit => 7-bit conversion. */
  rfc1522_encode_adrlist(en->to);
  rfc1522_encode_adrlist(en->cc);
  rfc1522_encode_adrlist(en->bcc);
  if (en->subject) {
    rfc1522_encode_string(buffer, (unsigned char *)en->subject, sizeof (buffer)-1);
    safe_free((void **)&en->subject);
    en->subject = safe_strdup(buffer);
  }
    }

  /* Write out the message in MIME form. */
  mutt_mktemp(tempfile);
  if ((tempfp = fopen (tempfile, "w")) == NULL) return (-1);

  mutt_write_rfc822_header(tempfp, en, attach, 0);
  fputc('\n', tempfp); /* tie off the header. */

  mutt_write_mime_body(attach, tempfp);
  fclose(tempfp);

  /* save a copy of the message, if necessary. */
  if (*fcc) mutt_folder_cc(tempfile, fcc);

  if (!batchmode) mutt_error("Sending message...");
  /*
   * invoke sendmail in the background so the user doesn't have to wait for
   * it to complete.  An example would be for someone who has dial-on-demand
   * networking.
   */
#ifdef USE_DSN
  sprintf(cmd, "%s -t -N %s -R %s", Sendmail, DsnNotify, DsnReturn);
#else
  sprintf(cmd, "%s -t", Sendmail);
#endif
#ifdef USE_8BITMIME
  if (attach->encoding == ENC8BIT)
    strcat(cmd, " -B8BITMIME");
#endif
  dprint(2,(debugfile,"mutt_send_message(): cmd=%s\n", cmd));
  sprintf(buffer, "(%s < %s; /bin/rm -f %s) &", cmd, tempfile, tempfile);
  mutt_system(buffer);
  if (!batchmode) mutt_error("Mail sent.");
  return 0;
}

void mutt_bounce_message (HEADER *h, char *to)
{
  int i;
  FILE *f;
  char buf[SHORT_STRING], date[SHORT_STRING];

  if (!h) {
    for (i=0; i<Context->msgcount; i++)
      if (Context->hdrs[i]->tagged)
	mutt_bounce_message(Context->hdrs[i], to);
    return;
  }

  fseek(MAILFP, h->offset, 0);
  snprintf(buf, sizeof(buf), "%s %s", Sendmail, to);

#ifdef USE_SETGID
  SETGID(User_gid);
#endif /* USE_SETGID */

  if ((f = popen(buf, "w")) != NULL) {
    mutt_copy_header(Context->folder->fp, h, f, CH_XMIT);
    fprintf(f, "Resent-From: %s", Username);
    if (Fqdn[0] != '@') fprintf(f, "@%s", Fqdn);
    fprintf(f, "\nResent-Date: %s", mutt_make_date(date));
    fprintf(f, "Resent-To: %s\n\n", to);
    mutt_copy_bytes(MAILFP, f, h->length - (ftell(MAILFP)-h->offset));
    pclose(f);
  }

#ifdef USE_SETGID
  SETGID(Mail_gid);
#endif /* USE_SETGID */
}

ADDRESS *mutt_default_from (void)
{
  ADDRESS *adr = mutt_new_address();

  if (Realname[0]) adr->personal = safe_strdup(Realname);
  adr->mailbox = safe_strdup(Username);
  adr->host = safe_strdup(Fqdn);
  return(adr);
}

/* given a list of addresses, return a list of unique addresses */
ADDRESS *mutt_remove_duplicates (ADDRESS *addr)
{
  ADDRESS *top = NULL;
  ADDRESS *tmp;
  
  if ((top = addr) == NULL) return(NULL);
  addr = addr->next;
  top->next = NULL;
  while (addr) {
    tmp = top;
    do {
      if (strcasecmp(addr->mailbox, tmp->mailbox) == 0 && strcasecmp(addr->host, tmp->host) == 0) {
	/* duplicate address, just ignore it */
	tmp = addr;
	addr = addr->next;
	tmp->next = NULL;
	mutt_free_address(&tmp);
      }
      else if (!tmp->next) {
	/* unique address.  add it to the list */
	tmp->next = addr;
	addr = addr->next;
	tmp = tmp->next;
	tmp->next = NULL;
	tmp = NULL; /* so we exit the loop */
      }
      else
	tmp = tmp->next;
    } while (tmp);
  }
  return(top);
}
